/* global console, define */
define(["./astro.core", "./astro.data", "./astro.math", "jquery"], function (core, data, math, $) {

	"use strict";

	// =========================================================================
	// packages
	// =========================================================================
	var calc = {};

	// =========================================================================
	// ecliptic
	// =========================================================================

	calc.eclipticAngle0 = math.deg2Rad(23.43928);

	calc.eclipticAngle = function (jd) {
		//return math.deg2Rad(23.5);
		var T = (jd - 2451545.0) / 36525.0;
		return math.deg2Rad(23.439291 - 0.013004 * T);
	};

	// =========================================================================
	// current date
	// =========================================================================
	calc.currentDate = function () {
		var d = new Date();
		return {
			day: d.getUTCDate(),
			month: d.getUTCMonth() + 1,
			year: d.getUTCFullYear(),
			hour: d.getUTCHours(),
			minute: d.getUTCMinutes()
		};
	};

	// =========================================================================
	// eccentric anomaly
	// =========================================================================
	calc.eccentricAnomaly = function (meanAnomaly, eccentricity) {
		var eccentricAnomaly = meanAnomaly + eccentricity * Math.sin(meanAnomaly);
		var deltaEccentricAnomaly = 1;
		while (Math.abs(deltaEccentricAnomaly) > 0.0000001) {
			deltaEccentricAnomaly = (meanAnomaly - eccentricAnomaly + eccentricity * Math.sin(eccentricAnomaly)) / (1 - eccentricity * Math.cos(eccentricAnomaly));
			eccentricAnomaly += deltaEccentricAnomaly;
		}
		return eccentricAnomaly;
	};

	// =========================================================================
	// orbital coordinate
	// =========================================================================
	calc.orbitalCoordinates = function (semiMajorAxis, eccentricity, eccentricAnomaly) {
		var lambda = 2 * Math.atan(Math.sqrt((1 + eccentricity) / (1 - eccentricity)) * Math.tan(eccentricAnomaly / 2));
		var rad = semiMajorAxis * (1 - eccentricity * Math.cos(eccentricAnomaly));
		return {
			lambda: lambda,
			rad: rad
		};
	};

	// =========================================================================
	// sidereal time
	// =========================================================================

	calc.siderealTime = function (jd, geoLongitude) {

		// montenbruck
		var st = 6.656306 + 24.0657098242 * (jd - 2445700.5);
		st = st * Math.PI / 12;
		if (geoLongitude) {
			st += geoLongitude;
		}
		return math.limit02Pi(st);
	};

	calc.siderealTime2 = function (jd) {

		// wiki

		// split into (jd at 0h UT) + UT
		var jd0 = Math.floor(jd - 0, 5) + 0.5;
		var ut = (jd - jd0) * 24;

		// st for 0h ut 
		var T = (jd0 - 2451545) / 36525;
		var st = 100.46061837 + 36000.770053608 * T + 0.000387933 * T * T - (T * T * T / 38710000);

		// add ut
		st += ut * 1.00273790935 * 360 / 24;

		return st * Math.PI / 180;
	};

	calc.siderealTime3 = function (jd) {

		// rough estimation

		var jd0 = calc.julianDate(21, 3, 1979, 12);
		var b = Math.PI * 2 * (1 + 1 / 365.2425);
		return (jd - jd0) * b;
	};

	// =========================================================================
	// julian date
	// =========================================================================

	$.extend(calc, {

		julianDate: function (day, month, year, ut) {
			var jd = null;
			if (month <= 2) {
				jd = calc.jd0(day, month, year);
			} else {
				jd = calc.jd0(day, month, year + 1) - 365;
			}
			jd += calc.daysUpToMonth(month) + day - 1;
			if (calc.isGregorian(day, month, year))
				jd += -10;
			jd += (ut - 12) / 24.0;
			return jd;
		},

		LEAPDAYDELTA: math.int(1582 / 100) - math.int(1582 / 400),

		jd0: function (day, month, year) {
			var result = math.int((4712 + year) * 365.25 + 0.75);
			if (calc.isGregorian(day, month, year)) {
				result += calc.LEAPDAYDELTA - math.int(year / 100) + math.int(year / 400);
			}
			return result;
		},

		isGregorian: function (day, month, year) {
			if (year < 1582)
				return false;
			if (year > 1582)
				return true;
			if (month < 10)
				return false;
			if (month > 10)
				return true;
			if (day <= 4)
				return false;
			return true;
		},

		daysUpToMonth: function (month) {
			var result = 0;
			for (var m = 1; m < month; m++) {
				result += calc.daysOfMonth(m);
			}
			return result;
		},

		daysOfMonth: function (month) {
			switch (month) {
			case 1:
				return 31;
			case 2:
				return 28;
			case 3:
				return 31;
			case 4:
				return 30;
			case 5:
				return 31;
			case 6:
				return 30;
			case 7:
				return 31;
			case 8:
				return 31;
			case 9:
				return 30;
			case 10:
				return 31;
			case 11:
				return 30;
			case 12:
				return 31;
			default:
				throw ("wrong month " + month);
			}

		},

		nameOfMonth: function (month) {
			switch (month) {
			case 1:
				return "January";
			case 2:
				return "February";
			case 3:
				return "March";
			case 4:
				return "April";
			case 5:
				return "May";
			case 6:
				return "June";
			case 7:
				return "Juli";
			case 8:
				return "August";
			case 9:
				return "September";
			case 10:
				return "October";
			case 11:
				return "November";
			case 12:
				return "December";
			default:
				throw ("wrong month " + month);
			}
		}

	});

	// =========================================================================
	// orbital elements of a planet
	// =========================================================================

	calc.orbitalElements = core.createClass({

		init: function (data, julianDate) {
			this.data = data;
			if (arguments.length === 2) {
				this.setJulianDate(julianDate);
			}
		},

		setJulianDate: function (julianDate) {
			this.julianDate = julianDate;
			var T = (julianDate - 2451545) / 36525.0;
			this.T = T;
			this.name = this.data.name;
			this.id = this.data.id;
			this.semiMajorAxis = this.calculateOE('semiMajorAxis', T);
			this.eccentricity = this.calculateOE('eccentricity', T);
			this.inclination = math.deg2Rad(this.calculateOE('inclination', T));
			this.meanLongitude = math.deg2Rad(this.calculateOE('meanLongitude', T));
			this.longitudePerihelion = math.deg2Rad(this.calculateOE('longitudePerihelion', T));
			this.longitudeAscendingNode = math.deg2Rad(this.calculateOE('longitudeAscendingNode', T));
			this.meanAnomaly = this.meanLongitude - this.longitudePerihelion;
			this.argumentPerihelion = this.longitudePerihelion - this.longitudeAscendingNode;
		},

		calculateOE: function (name, T) {
			if (this.data[name] === undefined) {
				return 0;
			}
			var offset = this.data[name][0];
			var delta = this.data[name][1];
			return offset + T * delta;
		}
	});

	// =========================================================================
	// list of all planets
	// =========================================================================    
	calc.allPlanets = (function () {
		var allPlanets = [];
		for (var id in data) {
			allPlanets.push(id);
		}
		return allPlanets;
	})();

	// =========================================================================
	// planets
	// =========================================================================
	calc.planets = core.createClass({

		allPlanets: calc.allPlanets,

		init: function (julianDate, geoLatitude, geoLongitude, ids) {

			// init
			this.julianDate = julianDate;
			this.geoLatitude = geoLatitude || math.rad2Deg(52);
			this.geoLongitude = geoLongitude || math.deg2Rad(13);
			this.ids = ids || this.allPlanets;
			this.siderealTime = calc.siderealTime(julianDate, geoLongitude);

			// calculate sun & earth
			this.list = [];
			this.earth = calc.planet(calc.orbitalElements(data.earth, this.julianDate), this);
			this.sun = calc.sun(calc.orbitalElements(data.sun, this.julianDate), this);
			this.list.push(this.sun, this.earth);

			// calculate remaining planets
			for (var i = 0; i < this.ids.length; ++i) {
				var id = this.ids[i];
				if (id === 'earth' || id === 'sun') {
					continue;
				}
				if (!data[id]) {
					continue;
				}
				this[id] = calc.planet(calc.orbitalElements(data[id], this.julianDate), this);
				this.list.push(this[id]);
			}
		}

	});

	// =========================================================================
	// planet
	// =========================================================================

	calc.planet = core.createClass({

		init: function (orbitalElements, planets) {

			this.orbitalElements = orbitalElements;
			this.planets = planets;

			this.calcHeliocentricOrbital();
			this.calcHeliocentricEcliptical();
			this.calcGeocentricEcliptical();
			this.calcGeocentricEquatorial();
			this.calcHorizontal();

			this.calcHeliocentricEclipticalApprox();
			this.calcGeocentricEclipticalApprox();
			this.calcGeocentricEquatorialApprox();
			this.calcHorizontalApprox();

			this.calcElongation();
			this.calcMagnitude();
		},

		calcHeliocentricOrbital: function () {
			var eccentricAnomaly = calc.eccentricAnomaly(this.orbitalElements.meanAnomaly, this.orbitalElements.eccentricity);
			var lambda = 2 * Math.atan(Math.sqrt((1 + this.orbitalElements.eccentricity) / (1 - this.orbitalElements.eccentricity)) * Math.tan(eccentricAnomaly / 2));
			var rad = this.orbitalElements.semiMajorAxis * (1 - this.orbitalElements.eccentricity * Math.cos(eccentricAnomaly));
			this.heliocentricOrbital = math.vector(0, lambda, rad, 'spherical');
		},

		calcHeliocentricEcliptical: function () {
			this.heliocentricEcliptical = this.heliocentricOrbital
				.rotZ(this.orbitalElements.argumentPerihelion)
				.rotX(this.orbitalElements.inclination)
				.rotZ(this.orbitalElements.longitudeAscendingNode);
			this.plot_helioX = this.heliocentricEcliptical.x;
			this.plot_helioY = this.heliocentricEcliptical.y;
			this.plot_dist_sun = this.heliocentricEcliptical.getDistance();
		},

		calcGeocentricEcliptical: function () {
			if (this.orbitalElements.data.id !== 'earth') {
				this.geocentricEcliptical = this.heliocentricEcliptical.subtract(this.planets.earth.heliocentricEcliptical);
			} else {
				this.geocentricEcliptical = math.vector();
			}
			this.plot_geoX = this.geocentricEcliptical.x;
			this.plot_geoY = this.geocentricEcliptical.y;
			this.plot_dist_eart = this.geocentricEcliptical.getDistance();
		},

		calcGeocentricEquatorial: function () {
			this.geocentricEquatorial = this.geocentricEcliptical.rotX(calc.eclipticAngle(this.planets.julianDate));
			this.plot_dec = math.rad2Deg(this.geocentricEquatorial.getLatitude());
			this.plot_rec = math.rad2Deg(this.geocentricEquatorial.getLongitude());
		},

		calcHorizontal: function () {
			this.horizontalMath = this.geocentricEquatorial.rotZ(-this.planets.siderealTime).rotY(-Math.PI / 2 + this.planets.geoLatitude);
			// this is the mathematical coordination system, longitude starting in the south, positive to east !!
			// (astronomical longitude = azimuth = starting in the south, positive to west)
			this.horizontalSouth = this.horizontalMath.mirrorY();
			this.horizontalNorth = this.horizontalMath.mirrorY().rotZ(Math.PI);
		},

		calcHeliocentricEclipticalApprox: function () {
			this.heliocentricEclipticalApprox = math.vector(0, this.orbitalElements.meanLongitude, this.orbitalElements.semiMajorAxis, 'spherical');
		},

		calcGeocentricEclipticalApprox: function () {
			if (this.orbitalElements.data.id !== 'earth') {
				this.geocentricEclipticalApprox = this.heliocentricEclipticalApprox.subtract(this.planets.earth.heliocentricEclipticalApprox);
			} else {
				this.geocentricEclipticalApprox = math.vector();
			}
		},

		calcGeocentricEquatorialApprox: function () {
			this.geocentricEquatorialApprox = this.geocentricEclipticalApprox.rotX(calc.eclipticAngle(this.planets.julianDate));
		},

		calcHorizontalApprox: function () {
			this.horizontalApprox = this.geocentricEquatorialApprox.rotZ(-this.planets.siderealTime).rotY(-Math.PI / 2 + this.planets.geoLatitude);
			// this is the mathematical coordination system, longitude starting in the south, positive to east !!
			// (astronomical longitude = asimuth = starting in the south, positive to west)
		},

		calcElongation: function () {

			if (this.orbitalElements.data.id === 'earth' || this.orbitalElements.data.id === 'sun') {
				this.elongation = 0;
				this.elongation2 = 0;
				return;
			}

			this.elongation = math.limit02Pi(-this.planets.sun.geocentricEcliptical.getLongitude() + this.geocentricEcliptical.getLongitude());
			if (this.elongation > Math.PI) {
				this.elongation -= 2 * Math.PI;
			}
			this.plot_elongation = math.rad2Deg(this.elongation);

		},

		calcMagnitudeMontenbruck: function () {
			if (this.orbitalElements.data.id === 'earth' || this.orbitalElements.data.id === 'sun') {
				this.magnitude = 0;
				return;
			}

			this.phaseAngle = this.heliocentricEcliptical.angle(this.heliocentricEcliptical.subtract(this.planets.earth.heliocentricEcliptical));

			var magnitude = this.orbitalElements.data.magnitude;
			if (magnitude === undefined) {
				this.magnitude = 0;
				return;
			}

			var phaseAngle2 = math.rad2Deg(this.phaseAngle) / 100.0;
			this.magnitude = magnitude[0] +
				magnitude[1] * phaseAngle2 +
				magnitude[2] * phaseAngle2 * phaseAngle2 +
				magnitude[3] * phaseAngle2 * phaseAngle2 * phaseAngle2 +
				5 * math.log10(this.heliocentricEcliptical.getDistance() * this.heliocentricEcliptical.subtract(this.planets.earth.heliocentricEcliptical).getDistance());
			if (this.orbitalElements.data.id === 'saturn') {
				var h1 = Math.sin(math.deg2Rad(52));
				this.magnitude += -2.6 * h1 + 1.25 * h1 * h1;
			}
		},

		calcMagnitude: function () {
			if (this.orbitalElements.data.id === 'earth' || this.orbitalElements.data.id === 'sun') {
				this.magnitude = 0;
				return;
			}

			this.phaseAngle = this.heliocentricEcliptical.angle(this.heliocentricEcliptical.subtract(this.planets.earth.heliocentricEcliptical));
			var phaseAngleDeg = math.rad2Deg(this.phaseAngle);

			var param = 5 * math.log10(this.heliocentricEcliptical.getDistance() * this.heliocentricEcliptical.subtract(this.planets.earth.heliocentricEcliptical).getDistance());
			var f1 = phaseAngleDeg / 100.0;
			var phase = phaseAngleDeg;

			switch (this.orbitalElements.data.id) {
			case 'mercury':
				if (phaseAngleDeg > 150.0) f1 = 1.5;
				this.magnitude = -0.36 + param + 3.8 * f1 - 2.73 * f1 * f1 + 2 * f1 * f1 * f1;
				break;
			case 'venus':
				this.magnitude = -4.29 + param + 0.09 * f1 + 2.39 * f1 * f1 - 0.65 * f1 * f1 * f1;
				break;
			case 'mars':
				this.magnitude = -1.52 + param + 0.016 * phase;
				break;
			case 'jupiter':
				this.magnitude = -9.25 + param + 0.005 * phase;
				break;
			case 'saturn':
				var T = this.orbitalElements.T;
				var a0 = (40.66 - 4.695 * T) * Math.PI / 180.0;
				var d0 = (83.52 + 0.403 * T) * Math.PI / 180.0;
				var cosDec = Math.cos(this.geocentricEquatorial.getLatitude());
				var sinDec = Math.sin(this.geocentricEquatorial.getLatitude());
				var sinx = -Math.cos(d0) * cosDec * Math.cos(a0 - this.geocentricEquatorial.getLongitude());
				sinx = Math.abs(sinx - Math.sin(d0) * sinDec);
				var rings = -2.6 * sinx + 1.25 * sinx * sinx;
				this.magnitude = -8.88 + param + 0.044 * phase + rings;
				break;
			case 'uranus':
				this.magnitude = -7.19 + param + 0.0028 * phase;
				break;
			case 'neptune':
				this.magnitude = -6.87 + param;
				break;
			}

			this.plot_magnitude = this.magnitude;
			this.plot_phaseAngle = phaseAngleDeg;

			// KStars
			//              double param = 5 * log10(rsun() * rearth() );
			//    double phase = this->phase().Degrees();
			//    double f1 = phase/100.;
			//
			//    if( name() == i18n( "Mercury" ) ) {
			//        if ( phase > 150. ) f1 = 1.5;
			//        magnitude = -0.36 + param + 3.8*f1 - 2.73*f1*f1 + 2*f1*f1*f1;
			//    } else if( name() == i18n( "Venus" ) ) {
			//        magnitude = -4.29 + param + 0.09*f1 + 2.39*f1*f1 - 0.65*f1*f1*f1;
			//    } else if( name() == i18n( "Mars" ) ) {
			//        magnitude = -1.52 + param + 0.016*phase;
			//    } else if( name() == i18n( "Jupiter" ) ) {
			//        magnitude = -9.25 + param + 0.005*phase;
			//    } else if( name() == i18n( "Saturn" ) ) {
			//        double T = num->julianCenturies();
			//        double a0 = (40.66-4.695*T)* dms::PI / 180.;
			//        double d0 = (83.52+0.403*T)* dms::PI / 180.;
			//        double sinx = -cos(d0)*cosDec*cos(a0 - ra().radians());
			//        sinx = fabs(sinx-sin(d0)*sinDec);
			//        double rings = -2.6*sinx + 1.25*sinx*sinx;
			//        magnitude = -8.88 + param + 0.044*phase + rings;
			//    } else if( name() == i18n( "Uranus" ) ) {
			//        magnitude = -7.19 + param + 0.0028*phase;
			//    } else if( name() == i18n( "Neptune" ) ) {
			//        magnitude = -6.87 + param;
			//    }

		},

		log: function () {
			console.log(this.orbitalElements.name);
			console.log(this.heliocentricOrbital.format('spherical-degrees'));
			console.log(this.heliocentricEcliptical.format('spherical-degrees'));
			console.log(this.geocentricEcliptical.format('spherical-degrees'));
			console.log(this.geocentricEquatorial.format('spherical-hours'));
		}
	});

	// =========================================================================
	// sun
	// =========================================================================

	calc.sun = core.createDerivedClass(calc.planet, {

		calcHeliocentricOrbital: function () {
			this.heliocentricOrbital = math.vector();
		},

		calcHeliocentricEcliptical: function () {
			this.heliocentricEcliptical = math.vector();
		}

	});

	return calc;
});