/* global console, define, window, $, document */
define(["./astro.core", "./astro.data", "./astro.math", "./astro.calc", "./astro.map", "jquery", "raphael", "flot", "flot.time", "flot.axislabels"], function (core, data, math, calc, map, $, raphael, flot, flottime, flotaxislables) {

    "use strict";

    var ColorMap = core.createClass({

        init: function () {
            this.colors = ['red', 'green', 'blue', 'yellow', 'orange'];
            this.colorIndex = 0;
            this.colorMap = {};
        },

        getColor: function (id) {
            var color = this.colorMap[id];
            if (color) {
                return color;
            }
            color = this.colors[this.colorIndex];
            this.colorMap[id] = color;
            this.colorIndex++;
            if (this.colorIndex >= this.colors.length) {
                this.colorIndex = 0;
            }
        }
    });

    var module = {

        generateTimeSlice: function (day, month, year, ut, geoLongitude) {
            var julianDate = calc.julianDate(day, month, year, ut);
            var text = '' + year + '/' + month + '/' + day + math.formatAngle('hours', ut * Math.PI / 12.0);
            return {
                julianDate: julianDate,
                siderealTime: calc.siderealTime(julianDate, geoLongitude),
                title: text
            };
        },

        generateTimeSlices: function (options) {

            var results = [];
            var year, ut, i, j;
            switch (options.timeMode) {
            case 'year/day':
                for (i = 1; i <= 12; ++i) {
                    for (j = 1; j <= calc.daysOfMonth(i); ++j) {
                        results.push(this.generateTimeSlice(j, i, options.year, options.ut, options.geoLongitude));
                    }
                }
                break;
            case 'year/month':
                for (i = 1; i <= 12; ++i) {
                    results.push(this.generateTimeSlice(1, i, options.year, options.ut, options.geoLongitude));
                }
                break;
            case 'year/month2':
                for (i = 1; i <= 12; ++i) {
                    results.push(this.generateTimeSlice(1, i, options.year, options.ut, options.geoLongitude));
                    results.push(this.generateTimeSlice(15, i, options.year, options.u, options.geoLongitudet));
                }
                break;
            case 'year/quarter':
                results.push(this.generateTimeSlice(1, 1, options.year, options.ut, options.geoLongitude));
                results.push(this.generateTimeSlice(1, 3, options.year, options.ut, options.geoLongitude));
                results.push(this.generateTimeSlice(1, 6, options.year, options.ut, options.geoLongitude));
                results.push(this.generateTimeSlice(1, 9, options.year, options.ut, options.geoLongitude));
                break;
            case 'day/hour':
                for (i = 0; i < 23; ++i) {
                    results.push(this.generateTimeSlice(options.day, options.month, options.year, i + 0.0000001, options.geoLongitude));
                }
                break;
            }
            return results;
        },

        drawTimePlot: function (drawArea) {

            drawArea.height('600px');

            var plot = [];
            var planets;
            var startDate = calc.julianDate(1, 1, 2016, 0);
            var startDateJs = Date.UTC(2016, 0, 1, 0, 0, 0);
            for (var date = startDate, dateJs = startDateJs; date < startDate + 365; ++date, dateJs += 24 * 60 * 60 * 1000) {
                planets = calc.planets(date);
                plot.push([dateJs, math.rad2Deg(planets.venus.heliocentricEcliptical.getLongitude())]);
            }

            $.plot(drawArea, [plot], {
                series: {
                    lines: {
                        show: true
                    },
                    points: {
                        show: true
                    }
                },
                xaxis: {
                    mode: "time",
                    minTickSize: [1, "month"],
                    min: (new Date(2016, 0, 1)).getTime(),
                    max: (new Date(2016, 11, 31)).getTime()
                },
                yaxis: {
                    min: 0,
                    max: 360,
                    axisLabel: 'Wahrscheinlichkeit in %',
                    axisLabelPadding: 10
                }
            });


        },

        createCheckBox: function (parentNode, label, value) {

            var containerNode = document.createElement('tr');
            parentNode.appendChild(containerNode);

            var checkBox = document.createElement('input');
            containerNode.appendChild(checkBox);
            checkBox.setAttribute('type', 'checkbox');
            if (value) {
                checkBox.setAttribute('checked', true);
            }

            containerNode.appendChild(document.createTextNode(label));

            return checkBox;
        },

        createDropwdown: function (parentNode, values) {

            var containerNode = document.createElement('tr');
            parentNode.appendChild(containerNode);

            var selectNode = document.createElement('select');
            containerNode.appendChild(selectNode);
            for (var i = 0; i < values.length; ++i) {
                var value = values[i];
                var optionNode = document.createElement('option');
                selectNode.appendChild(optionNode);
                optionNode.appendChild(document.createTextNode(value));
                optionNode.setAttribute('value', value);
            }
            return selectNode;
        },

        createInputField: function (parentNode, label, value, comment) {

            // container
            var containerNode = document.createElement('tr');
            parentNode.appendChild(containerNode);

            // label
            var labelNode = document.createElement('td');
            containerNode.appendChild(labelNode);
            labelNode.appendChild(document.createTextNode(label));

            // value
            var valueNode = document.createElement('td');
            containerNode.appendChild(valueNode);
            var inputNode = document.createElement('input');
            if (value) {
                inputNode.setAttribute('value', value);
            }
            valueNode.appendChild(inputNode);

            // comment
            if (comment) {
                var commentNode = document.createElement('td');
                commentNode.appendChild(document.createTextNode(comment));
                containerNode.appendChild(commentNode);
            }

            return inputNode;
        },

        drawStarMap: function (options) {

            var i, time, siderealTime;

            var colorMap = new ColorMap();

            // fill default values
            options = options || {};
            options.objects = options.objects || ['sun', 'saturn'];
            options.year = options.year || (new Date()).getFullYear();
            options.geoLatitude = options.geoLatitude || math.deg2Rad(52);
            options.geoLongitude = options.geoLongitude || math.deg2Rad(13);
            options.type = options.type || 'sky';
            options.timeMode = options.timeMode || 'year/month';

            // draw empty star map
            var paper = raphael(options.node, 600, 600);
            var mymap = map.simpleMap({
                paper: paper
            });
            mymap.draw(['equator']);

            // generate time
            var times = module.generateTimeSlices(options);

            // planets 
            var planets;
            for (i = 0; i < times.length; ++i) {
                time = times[i];
                switch (options.mapMode) {
                case 'starmap':
                    break;
                case 'sky':
                    mymap.setGeoLocation(time.siderealTime, options.geoLatitude);
                    break;
                }
                planets = calc.planets(time.julianDate, options.geoLatitude, options.geoLongitude, options.objects);
                for (var j = 0; j < planets.list.length; ++j) {
                    var planet = planets.list[j];
                    if (options.objects.indexOf(planet.orbitalElements.id) < 0) {
                        continue;
                    }
                    if (options.timeMode === 'day/hour' && options.mapMode === 'sky') {
                        planet.geocentricEquatorial.title = i;
                    }
                    planet.geocentricEquatorial.tooltip = planet.orbitalElements.name + '\n' + time.title;
                    planet.geocentricEquatorial.color = colorMap.getColor(planet.orbitalElements.id);
                    mymap.drawPlanet(planet.geocentricEquatorial);
                }
            }

            // ecliptic and stars          
            if (options.objects.indexOf('ecliptic') >= 0 || options.objects.indexOf('stars') >= 0) {
                switch (options.mapMode) {
                case 'starmap':
                    if (options.objects.indexOf('ecliptic') >= 0) {
                        mymap.drawEcliptic();
                    }
                    if (options.objects.indexOf('stars') >= 0) {
                        mymap.drawStarLines();
                        mymap.drawStars();
                    }
                    break;
                case 'sky':
                    for (i = 0; i < times.length; ++i) {
                        time = times[i];
                        mymap.setGeoLocation(time.siderealTime, options.geoLatitude);
                        if (options.objects.indexOf('ecliptic') >= 0) {
                            mymap.drawEcliptic();
                        }
                        if (options.objects.indexOf('stars') >= 0) {
                            mymap.drawStarLines();
                            mymap.drawStars();
                        }
                    }
                    break;
                }
            }

        },

        calculate: function () {

            var year = parseInt(this.yearInputNode.value, 10);
            var month = parseInt(this.monthInputNode.value, 10);
            var day = parseInt(this.dayInputNode.value, 10);
            var hour = parseInt(this.hourInputNode.value, 10);
            var minute = parseInt(this.minuteInputNode.value, 10);
            var ut = hour + minute / 60;

            var geoLatitude = parseFloat(this.geoLatitudeInputNode.value) * Math.PI / 180;
            var geoLongitude = parseFloat(this.geoLongitudeInputNode.value) * Math.PI / 180;

            var objects = [];
            for (var i = 0; i < this.objectCheckboxes.length; ++i) {
                var objectCheckbox = this.objectCheckboxes[i];
                if (objectCheckbox.checkBox.checked) {
                    objects.push(objectCheckbox.object);
                }
            }

            var mapMode = this.mapModeDropdown.value;
            var timeMode = this.timeModeDropdown.value;

            this.mapArea.innerHTML = '';
            this.drawStarMap({
                timeMode: timeMode,
                objects: objects,
                year: year,
                month: month,
                day: day,
                ut: ut,
                geoLatitude: geoLatitude,
                geoLongitude: geoLongitude,
                mapMode: mapMode,
                node: this.mapArea
            });

        },

        draw: function () {

            var self = this;
            this.rootNode = document.getElementById('area');

            this.inputTableNode = document.createElement('table');
            this.rootNode.appendChild(this.inputTableNode);

            var date = calc.currentDate();
            this.dayInputNode = this.createInputField(this.inputTableNode, "day", "" + date.day, "UTC (1..31)");
            this.monthInputNode = this.createInputField(this.inputTableNode, "month", "" + date.month, "UTC (1..12)");
            this.yearInputNode = this.createInputField(this.inputTableNode, "year", "" + date.year, "UTC");
            this.hourInputNode = this.createInputField(this.inputTableNode, "hour", "" + date.hour, "UTC (0..23, UTC=MEZ-1 or UTC=MESZ-2)");
            this.minuteInputNode = this.createInputField(this.inputTableNode, "minute", "" + date.minute, "UTC (0..59)");

            // geo input fields
            this.geoLatitudeInputNode = this.createInputField(this.inputTableNode, "latitude", 52, "degrees (berlin = 52 )");
            this.geoLongitudeInputNode = this.createInputField(this.inputTableNode, "longitude", 13, "degrees (<0 for west, berlin = 13)");

            // object checkboxes
            var objects = calc.allPlanets.slice();
            objects.push('ecliptic', 'stars');
            this.objectCheckboxes = [];
            for (var i = 0; i < objects.length; ++i) {
                var object = objects[i];
                var checkbox = this.createCheckBox(this.inputTableNode, object, false);
                this.objectCheckboxes.push({
                    checkBox: checkbox,
                    object: object
                });
            }

            // mapMode dropdown
            this.mapModeDropdown = this.createDropwdown(this.inputTableNode, ['starmap', 'sky']);

            // time mode dropwon
            this.timeModeDropdown = this.createDropwdown(this.inputTableNode, ['year/month', 'year/month2', 'year/quarter', 'year/day', 'day/hour']);

            // calculate button
            this.calculateButtonNode = document.createElement('button');
            this.calculateButtonNode.appendChild(document.createTextNode('Draw'));
            this.calculateButtonNode.addEventListener('click', function () {
                self.calculate();
            });
            this.rootNode.appendChild(this.calculateButtonNode);

            // map area
            this.mapArea = document.createElement('div');
            this.rootNode.appendChild(this.mapArea);
        }

    };
    return module;

});