/* global console, define, document */
define(["./astro.core", "./astro.math", "./astro.calc", "./astro.data","jquery"], function (core, math, calc, data,$) {

    "use strict";

    var printer = new core.Printer(document.body);
    var test = {};
    
    // =========================================================================
    // test helpers
    // =========================================================================

    $.extend(test, {

        error: function (msg) {
            printer.print("Error: " + msg);
        },

        info: function (msg) {
            printer.print("Info: " + msg);
        },

        checkFloat: function (name, calculated, expected, precision) {
            if (arguments.length < 4) {
                precision = 0;
            }
            if (Math.abs(calculated - expected) > precision) {
                test.error(name + " Expected: " + expected + " Got: " + calculated);
                return false;
            }
            return true;
        },

        checkVector: function (name, calculated, expected, precision) {
            if (arguments.length < 4) {
                precision = 0;
            }
            if (Math.abs(calculated.x - expected.x) > precision || Math.abs(calculated.y - expected.y) > precision || Math.abs(calculated.z - expected.z) > precision) {
                test.error(name + " Expected:" + expected + " Got:" + calculated);
            }
        },

        checkJD: function (name, day, month, year, jd) {
            var t = calc.julianDate(day, month, year, 12);
            if (Math.abs(t - jd) > 0) {
                test.error(name + " Expected: " + jd + " got: " + t);
            }
        },

        checkString: function (name, value, expected) {
            if (value !== expected) {
                test.error(name + " Expected :" + expected + " got :" + value);
            }
        }
    });

    // =========================================================================
    // tests
    // =========================================================================

    $.extend(test, {

        testSphericalCoordinates: function () {

            // test longitude
            test.checkFloat("longitude for (1,0,0)", math.rad2Deg(math.vector(1, 0, 0).getLongitude()), 0);
            test.checkFloat("longitude for (1,1,0)", math.rad2Deg(math.vector(1, 1, 0).getLongitude()), 45);
            test.checkFloat("longitude for (0,1,0)", math.rad2Deg(math.vector(0, 1, 0).getLongitude()), 90);
            test.checkFloat("longitude for (-1,1,0)", math.rad2Deg(math.vector(-1, 1, 0).getLongitude()), 135);
            test.checkFloat("longitude for (-1,0,0)", math.rad2Deg(math.vector(-1, 0, 0).getLongitude()), 180);
            test.checkFloat("longitude for (-1,-1,0)", math.rad2Deg(math.vector(-1, -1, 0).getLongitude()), 225);
            test.checkFloat("longitude for (0,-1,0)", math.rad2Deg(math.vector(0, -1, 0).getLongitude()), 270);
            test.checkFloat("longitude for (1,-1,0)", math.rad2Deg(math.vector(1, -1, 0).getLongitude()), 315);

            // test latitude
            test.checkFloat("longitude for (0,0,1)", math.rad2Deg(math.vector(0, 0, 1).getLatitude()), 90);
            test.checkFloat("longitude for (0,0,0)", math.rad2Deg(math.vector(0, 0, 0).getLatitude()), 0);
            test.checkFloat("longitude for (0,0,-1)", math.rad2Deg(math.vector(0, 0, -1).getLatitude()), -90);

        },

        testVector: function () {
            test.checkVector("add (1,1,1)+(1,2,3)", math.vector(1, 1, 1).add(math.vector(1, 2, 3)), math.vector(2, 3, 4));
            test.checkVector("subtract (5,5,5)-(1,2,3)", math.vector(5, 5, 5).subtract(math.vector(1, 2, 3)), math.vector(4, 3, 2));
            test.checkVector("multiplication 3*(1,2,3)", math.vector(1, 2, 3).multiply(3), math.vector(3, 6, 9));
            test.checkFloat("scalar multiplication (3,2,1)*(1,2,3)", math.vector(3, 2, 1).scalarProduct(math.vector(1, 2, 3)), 10);
        },

        testVectorRotation: function () {
            test.checkVector("z rotation 90deg (1,1,1)", (math.vector(1, 1, 1)).rotZ(math.deg2Rad(90)), math.vector(-1, 1, 1), 0.01);
            test.checkVector("x rotation 90deg (1,1,1)", (math.vector(1, 1, 1)).rotX(math.deg2Rad(90)), math.vector(1, -1, 1), 0.01);
            test.checkVector("y rotation 90deg (1,1,1)", (math.vector(1, 1, 1)).rotY(math.deg2Rad(90)), math.vector(1, 1, -1), 0.01);
        },

        testJD: function () {

            test.checkJD("19.6.2006", 19, 6, 2006, 2453906);
            test.checkJD("20.6.2006", 20, 6, 2006, 2453907);
            test.checkJD("20.6.2005", 20, 6, 2005, 2453542);
            test.checkJD("20.6.2004", 20, 6, 2004, 2453177);
            test.checkJD("20.6.2000", 20, 6, 2000, 2451716);
            test.checkJD("20.6.2000", 20, 6, 1999, 2451350);
            test.checkJD("20.6.1983", 20, 6, 1983, 2445506);
            test.checkJD("20.6.1980", 20, 6, 1980, 2444411);

            test.checkJD("20.1.1980", 20, 1, 1980, 2444259);
            test.checkJD("20.2.1980", 20, 2, 1980, 2444290);
            test.checkJD("20.3.1980", 20, 3, 1980, 2444319);
            test.checkJD("20.4.1980", 20, 4, 1980, 2444350);
            test.checkJD("20.5.1980", 20, 5, 1980, 2444380);
            test.checkJD("20.6.1980", 20, 6, 1980, 2444411);

            test.checkJD("1.1.-4712", 1, 1, -4712, 0);
            test.checkJD("1.1.-4711", 1, 1, -4711, 366);
            test.checkJD("1.1.-4710", 1, 1, -4710, 731);
            test.checkJD("1.1.-4709", 1, 1, -4709, 1096);
            test.checkJD("1.1.-4708", 1, 1, -4708, 1461);
            test.checkJD("1.1.-4707", 1, 1, -4707, 1827);

            test.checkJD("1.2.-4712", 1, 2, -4712, 31);
            test.checkJD("1.3.-4712", 1, 3, -4712, 60);

            test.checkJD("1.2.-4711", 1, 2, -4711, 397);
            test.checkJD("1.3.-4711", 1, 3, -4711, 425);

            test.checkJD("1.2.-4710", 1, 2, -4710, 762);
            test.checkJD("1.3.-4710", 1, 3, -4710, 790);

            test.checkJD("1.2.-4709", 1, 2, -4709, 1127);
            test.checkJD("1.3.-4709", 1, 3, -4709, 1155);

            test.checkJD("1.2.-4708", 1, 2, -4708, 1492);
            test.checkJD("1.3.-4708", 1, 3, -4708, 1521);

            test.checkJD("1.2.-4707", 1, 2, -4707, 1858);
            test.checkJD("1.3.-4707", 1, 3, -4707, 1886);

            // julianisch -> gregorianisch
            test.checkJD("1.2.1582", 1, 2, 1582, 2298915);
            test.checkJD("1.3.1582", 1, 3, 1582, 2298943);
            test.checkJD("4.10.1582", 4, 10, 1582, 2299160);
            test.checkJD("15.10.1582", 15, 10, 1582, 2299161);
            test.checkJD("1.2.1583", 1, 2, 1583, 2299270);
            test.checkJD("1.3.1583", 1, 3, 1583, 2299298);

            // 2000
            test.checkJD("1.2.2000", 1, 2, 2000, 2451576);
            test.checkJD("1.3.2000", 1, 3, 2000, 2451605);

        },

        testKeplerEquation: function () {
            var eccentricity = 0.047837;
            var meanAnomaly = math.deg2Rad(199.7855);
            var eccentricAnomlay = calc.eccentricAnomaly(meanAnomaly, eccentricity);
            test.checkFloat("kepler equation ", math.rad2Deg(eccentricAnomlay), 198.89778, 0.00001);
        },

        testAngle: function () {
            var angle = math.parseAngle("hours", "12:00:00");
            test.checkString("hour parsing 12:00:00 ", math.formatAngle("degrees", angle), "+180:00:00");
            angle = math.parseAngle("degrees", "180:00:00");
            test.checkString("degree parsing 12:00:00 ", math.formatAngle("degrees", angle), "+180:00:00");
        },

        sideralTime: function () {
            printer.print("sideral");
            var jd0 = calc.julianDate(22, 3, 1972, 12);
            var jd1 = calc.julianDate(1, 1, 1982, 0);

            var b = 24 * (1 + 1 / 365.2425);
            var st = b * (jd1 - jd0);
            printer.print("st=" + math.formatAngle("hours", st * Math.PI * 2 / 24));

            var a = 6.656306;
            var c = (6.656306 - 12 + b * jd0) / (b - 24);
            var dc = 24 / (b - 24);
            c -= dc;
            while (c > 2445701.5) {
                c -= dc;
            }
            // c=2445700.5;
            st = a + (b - 24) * (jd1 - c);
            printer.print("b-24=" + (b - 24));
            printer.print("c=" + c);
            printer.print("st=" + math.formatAngle("hours", st * Math.PI * 2 / 24));
        },

        sideralTime2: function () {
            var jd = calc.julianDate(1, 1, 1982, 0);
            printer.print("st1=" + math.formatAngle("hours", calc.siderealTime(jd)));
            printer.print("st2=" + math.formatAngle("hours", calc.siderealTime2(jd)));

            jd = calc.julianDate(22, 6, 2014, 16);
            printer.print("st1=" + math.formatAngle("hours", calc.siderealTime(jd)));
            printer.print("st2=" + math.formatAngle("hours", calc.siderealTime2(jd)));

            jd = calc.julianDate(1, 1, 2013, 12);
            var i = 0;
            for (var m = 1; m <= 12; ++m) {
                for (var d = 1; d <= calc.daysOfMonth(m); ++d) {
                    printer.print("" + d + "." + m + ":" + math.formatAngle("hours", calc.siderealTime(jd + i)));
                    i += 1;
                }
            }
        },

        calc: function () {
            var julianDate = calc.julianDate(3, 8, 2010, 19);
            var earth = calc.planet(calc.orbitalElements(data.earth, julianDate));
            var venus = calc.planet(calc.orbitalElements(data.venus, julianDate), earth);
            var sun = calc.sun(earth);
            venus.log();
            earth.log();
            sun.log();
        },

        ZGL: function () {

            var zgl = function (year) {
                var julianDate = calc.julianDate(1, 1, year, 12);
                var plotd = [];
                for (var d = 0; d < 365; ++d) {
                    var earth = calc.planet(calc.orbitalElements(data.earth, julianDate));
                    var sun = calc.sun(earth);
                    var st = calc.siderealTime(julianDate);
                    var diff = math.limit02Pi(st - sun.geocentricEquatorial.getLongitude());
                    if (diff > Math.PI) {
                        diff -= 2 * Math.PI;
                    }
                    diff *= 12 / Math.PI * 60;
                    plotd.push([d, diff]);
                    julianDate += 1;
                }
                return plotd;
            };


            $.plot($("#zgl"), [zgl(2012), zgl(2015)], {
                yaxis: {
                    max: 20,
                    min: -20
                },
                xaxis: {
                    max: 365,
                    min: 0
                }
            });
        },

        testAll: function () {

            test.testSphericalCoordinates();
            test.testVector();
            test.testVectorRotation();
            test.testJD();
            test.testKeplerEquation();
            test.testAngle();

            test.info("Finished!");
        }
    });

    test.testAll();

    return test;
});