(function(global) {"use strict";

	// =========================================================================
	// imported packages
	// =========================================================================
	var core = global.astro.core;
	var data = global.astro.data;
	var math = global.astro.math;
	var calc = global.astro.calc;
	var dia = global.astro.dia;
	var map = global.astro.map;

	var $ = global.$;

	// =========================================================================
	// exported packages
	// =========================================================================
	var eph = global.astro.eph = {};
	var module = eph;

	// =========================================================================
	// formatter for ports
	// =========================================================================
	module.angleFormatter = function(value) {
		return math.formatAngle('degrees-simple', value);
	};
	module.vectorFormatter = function(value) {
		if ( value instanceof math.vectorList) {
			return "list of vectors";
		}
		var v = math.vector(value.x, value.y, value.z);
		return v.format('spherical-degrees-simple');
	};

	// =========================================================================
	// vector aggregator
	// =========================================================================
	module.vectorAggregator = function() {
		var result = math.vectorList();
		for (var i = 0; i < arguments.length; ++i) {
			result = result.addVector(arguments[i]);
		}
		return result;
	};

	// =========================================================================
	// deserializator for vector
	// =========================================================================
	module.VectorDeserializator = {

		match : function(value) {
			if (value.x !== undefined && value.y !== undefined && value.z !== undefined) {
				return true;
			}
			return false;
		},

		deSerialize : function(value) {
			var v = math.vector(value.x, value.y, value.z);
			for (var property in value) {
				if (property === 'x' || property === 'y' || property === 'z') {
					continue;
				}
				v[property] = value[property];
			}
			return v;
		}
	};

	// =========================================================================
	// deserializator for vectorlist
	// =========================================================================
	module.VectorListDeserializator = {

		match : function(value) {
			if (value.vectors !== undefined) {
				return true;
			}
			return false;
		},

		deSerialize : function(value) {
			var vl = math.vectorList();
			for (var i = 0; i < value.vectors.length; ++i) {
				vl.vectors.push(module.VectorDeserializator.deSerialize(value.vectors[i]));
			}
			return vl;
		}
	};

	module.Deserializators = [module.VectorDeserializator, module.VectorListDeserializator];

	// =========================================================================
	// deserializator for values
	// =========================================================================
	dia.deserializeValue = function(value) {
		for (var i = 0; i < module.Deserializators.length; ++i) {
			var deserializator = module.Deserializators[i];
			if (deserializator.match(value)) {
				return deserializator.deSerialize(value);
			}
		}
		return value;
	};

	// =========================================================================
	// Julian date
	// =========================================================================
	module.JulianDate = dia.Element.extend({

		object : "JulianDate",
		module : "astro.eph",

		init : function(properties) {

			// base class init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Julian Date',
				value : {
					julianDate : calc.julianDate(1, 1, 2000, 0),
					day : 1,
					month : 1,
					year : 2000,
					ut : 0
				}
			}, properties);

			// ports
			this.setNumberPorts(dia.SOUTH, 1);
			this.integerOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'Julian date:',
				value : calc.julianDate(1, 1, 2000, 0),
				calculatePort : function() {
					return this.properties.value.julianDate;
				}
			});

		},

		text : function() {
			return "julianDate=" + this.properties.value.julianDate;
		},

		getProperties : function() {
			return [this.getTitleProperty(), {
				label : 'day',
				object : this.properties.value,
				name : 'day',
				convertFromUI : parseInt
			}, {
				label : 'month',
				object : this.properties.value,
				name : 'month',
				convertFromUI : parseInt
			}, {
				label : 'year',
				object : this.properties.value,
				name : 'year',
				convertFromUI : parseInt
			}, {
				label : 'ut',
				object : this.properties.value,
				name : 'ut',
				convertFromUI : parseFloat
			}];
		},

		endOfUpdate : function() {
			var value = this.properties.value;
			value.julianDate = calc.julianDate(value.day, value.month, value.year, value.ut);
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Vector
	// =========================================================================

	module.Vector = dia.Element.extend({

		object : "Vector",
		module : "astro.eph",

		init : function(properties) {

			// base class init
			var v = math.vector(0, 0, 0);
			v.mode = "spherical-degrees";
			this.baseInit({
				width : 160,
				height : 60,
				title : 'Vector',
				value : v
			}, properties);

			// ports
			this.setNumberPorts(dia.NORTH, 1);
			this.setNumberPorts(dia.SOUTH, 1);
			this.setNumberPorts(dia.WEST, 3);
			this.vectorInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'vector:',
				formatter : module.vectorFormatter,
				value : math.nullVector,
				defaultValue : math.nullVector,
				aggregator : module.vectorAggregator
			});
			this.betaInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'beta:',
				formatter : module.angleFormatter,
				value : 0,
				defaultValue : 0
			});
			this.lambdaInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'lambda:',
				formatter : module.angleFormatter,
				value : 0,
				defaultValue : 0
			});
			this.distanceInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'distance:',
				value : 1,
				defaultValue : 1
			});
			this.vectorOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'vector:',
				formatter : module.vectorFormatter,
				value : math.nullVector,
				calculatePort : function() {
					return this.properties.value;
				}
			});

		},

		calculate : function() {
			var mode = this.properties.value.mode;
			if (this.lambdaInPort.hasInput() || this.betaInPort.hasInput()) {
				this.properties.value = math.vector(this.betaInPort.value(), this.lambdaInPort.value(), this.distanceInPort.value(), "spherical");
			} else if (this.vectorInPort.hasInput()) {
				this.properties.value = math.vector(this.vectorInPort.value().x, this.vectorInPort.value().y, this.vectorInPort.value().z);
			}
			this.properties.value.mode = mode;
		},

		text : function() {
			return this.properties.value.format(this.properties.value.mode);
		},

		getProperties : function() {
			var self = this;
			var modeProperty = {
				label : 'mode',
				get : function() {
					return self.properties.value.mode;
				},
				set : function(value) {
					self.properties.value.mode = value;
					self.updateTextArea();
					self.refreshPropertyEditor();
				},
				immediateInput : true,
				type : 'dropdown',
				values : ['spherical-degrees', 'spherical-degrees-simple', 'spherical-hours', 'cartesian']
			};
			var readonly = this.vectorInPort.hasInput() || this.lambdaInPort.hasInput() || this.betaInPort.hasInput();
			switch(this.properties.value.mode) {
				case 'cartesian':
					this.userInput = {
						x : this.properties.value.x,
						y : this.properties.value.y,
						z : this.properties.value.z
					};
					return [modeProperty, this.getTitleProperty(), {
						label : 'x',
						object : this.userInput,
						name : 'x',
						readonly : readonly
					}, {
						label : 'y',
						object : this.userInput,
						name : 'y',
						readonly : readonly
					}, {
						label : 'z',
						object : this.userInput,
						name : 'z',
						readonly : readonly
					}];
				case 'spherical-degrees':
					this.userInput = {
						longitude : math.formatAngle('degrees', this.properties.value.getLongitude()),
						latitude : math.formatAngle('degrees-latitude', this.properties.value.getLatitude()),
						distance : this.properties.value.getDistance()
					};
					return [modeProperty, this.getTitleProperty(), {
						label : 'longitude (degrees)',
						object : this.userInput,
						name : 'longitude',
						readonly : readonly
					}, {
						label : 'latitude (degrees)',
						object : this.userInput,
						name : 'latitude',
						readonly : readonly
					}, {
						label : 'distance',
						object : this.userInput,
						name : 'distance',
						readonly : readonly
					}];
				case 'spherical-degrees-simple':
					var v = this.properties.value;
					this.userInput = {
						longitude : math.rad2Deg(v.getLongitude()).toFixed(2),
						latitude : math.rad2Deg(v.getLatitude()).toFixed(2),
						distance : v.getDistance()
					};
					return [modeProperty, this.getTitleProperty(), {
						label : 'longitude (degrees)',
						object : this.userInput,
						name : 'longitude',
						readonly : readonly
					}, {
						label : 'latitude (degrees)',
						object : this.userInput,
						name : 'latitude',
						readonly : readonly
					}, {
						label : 'distance',
						object : this.userInput,
						name : 'distance',
						readonly : readonly
					}];
				case 'spherical-hours':
					v = this.properties.value;
					this.userInput = {
						longitude : math.formatAngle('hours', v.getLongitude()),
						latitude : math.formatAngle('degrees-latitude', v.getLatitude()),
						distance : v.getDistance()
					};
					return [modeProperty, this.getTitleProperty(), {
						label : 'longitude (hours)',
						object : this.userInput,
						name : 'longitude',
						readonly : readonly
					}, {
						label : 'latitude (degrees)',
						object : this.userInput,
						name : 'latitude',
						readonly : readonly
					}, {
						label : 'distance',
						object : this.userInput,
						name : 'distance',
						readonly : readonly
					}];
			}
		},

		endOfUpdate : function() {
			switch(this.properties.value.mode) {
				case 'cartesian':
					this.properties.value.x = parseFloat(this.userInput.x);
					this.properties.value.y = parseFloat(this.userInput.y);
					this.properties.value.z = parseFloat(this.userInput.z);
					break;
				case 'spherical-degrees':
					var longitude = math.parseAngle("degrees", this.userInput.longitude);
					var latitude = math.parseAngle("degrees", this.userInput.latitude);
					var distance = parseFloat(this.userInput.distance);
					var v = math.vector(latitude, longitude, distance, "spherical");
					this.properties.value.x = v.x;
					this.properties.value.y = v.y;
					this.properties.value.z = v.z;
					break;
				case 'spherical-degrees-simple':
					longitude = math.deg2Rad(parseFloat(this.userInput.longitude));
					latitude = math.deg2Rad(parseFloat(this.userInput.latitude));
					distance = parseFloat(this.userInput.distance);
					v = math.vector(latitude, longitude, distance, "spherical");
					this.properties.value.x = v.x;
					this.properties.value.y = v.y;
					this.properties.value.z = v.z;
					break;
				case 'spherical-hours':
					longitude = math.parseAngle("hours", this.userInput.longitude);
					latitude = math.parseAngle("degrees", this.userInput.latitude);
					distance = parseFloat(this.userInput.distance);
					v = math.vector(latitude, longitude, distance, "spherical");
					this.properties.value.x = v.x;
					this.properties.value.y = v.y;
					this.properties.value.z = v.z;
					break;
			}
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Angle
	// =========================================================================

	module.Angle = dia.Element.extend({

		object : "Angle",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 160,
				height : 60,
				title : 'Angle',
				value : {
					angle : 0,
					mode : 'degrees'
				}
			}, properties);

			// ports
			this.setNumberPorts(dia.SOUTH, 1);
			this.setNumberPorts(dia.NORTH, 1);
			this.angleInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'angle:',
				formatter : module.angleFormatter,
				value : 0,
				defaultValue : 0
			});
			this.angleOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'angle:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.angle;
				}
			});

		},

		calculate : function() {
			if (this.angleInPort.hasInput()) {
				this.properties.value.angle = this.angleInPort.value();
			}
		},

		text : function() {
			return math.formatAngle(this.properties.value.mode, this.properties.value.angle);
		},

		getProperties : function() {
			var self = this;
			var modeProperty = {
				label : 'mode',
				get : function() {
					return self.properties.value.mode;
				},
				set : function(value) {
					self.properties.value.mode = value;
					self.updateTextArea();
					self.refreshPropertyEditor();
				},
				immediateInput : true,
				type : 'dropdown',
				values : ['degrees', 'degrees-simple', 'hours']
			};
			return [modeProperty, this.getTitleProperty(), {
				label : 'angle (' + this.properties.value.mode + ')',
				object : this.properties.value,
				name : 'angle',
				readonly : this.angleInPort.hasInput(),
				convertFromUI : function(value) {
					return math.parseAngle(self.properties.value.mode, value);
				},
				convertToUI : function(value) {
					return math.formatAngle(self.properties.value.mode, value);
				}
			}];

		},

		endOfUpdate : function() {
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Distance
	// =========================================================================

	module.Distance = dia.Element.extend({

		object : "Distance",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Distance',
				value : {
					distance : 0
				}
			}, properties);

			// ports
			this.setNumberPorts(dia.SOUTH, 1);
			this.setNumberPorts(dia.NORTH, 1);
			this.distanceInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'distance:',
				value : 0,
				defaultValue : 0
			});
			this.distanceOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'distance:',
				value : 0,
				calculatePort : function() {
					return this.properties.value.distance;
				}
			});

		},

		calculate : function() {
			if (this.distanceInPort.hasInput()) {
				this.properties.value.distance = this.distanceInPort.value();
			}
		},

		text : function() {
			return math.trimLength(5, this.properties.value.distance.toFixed(2));
		},

		getProperties : function() {
			var self = this;
			return [this.getTitleProperty(), {
				label : 'distance',
				object : this.properties.value,
				name : 'distance',
				readonly : this.distanceInPort.hasInput(),
				convertFromUI : parseFloat
			}];

		},

		endOfUpdate : function() {
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Rotation
	// =========================================================================

	module.Rotation = dia.Element.extend({

		object : "Rotation",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Rotation',
				value : {
					axis : 'x'
				}
			}, properties);

			// ports
			this.setNumberPorts(dia.SOUTH, 1);
			this.setNumberPorts(dia.NORTH, 1);
			this.setNumberPorts(dia.WEST, 1);
			this.angleInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'angle:',
				defaultValue : 0,
				value : 0,
				formatter : module.angleFormatter
			});
			this.vectorInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'vector:',
				value : math.nullVector,
				defaultValue : math.nullVector,
				formatter : module.vectorFormatter,
				aggregator : module.vectorAggregator
			});
			this.vectorOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'vector:',
				value : math.nullVector,
				formatter : module.vectorFormatter,
				calculatePort : function() {
					var v = this.vectorInPort.value();
					switch(this.properties.value.axis) {
						case 'x':
							v = v.rotX(this.angleInPort.value());
							break;
						case 'y':
							v = v.rotY(this.angleInPort.value());
							break;
						case 'z':
							v = v.rotZ(this.angleInPort.value());
							break;
					}
					return v;
				}
			});

		},

		text : function() {
			return this.properties.value.axis;
		},

		getProperties : function() {
			var self = this;
			return [this.getTitleProperty(), {
				label : 'axis',
				object : this.properties.value,
				name : 'axis',
				type : 'dropdown',
				values : ['x', 'y', 'z']
			}];

		},

		endOfUpdate : function() {
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Addition
	// =========================================================================

	module.Addition = dia.Element.extend({

		object : "Addition",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Addition',
				value : {
					angle : 0,
					operator : '+'
				}
			}, properties);

			// ports
			this.setNumberPorts(dia.NORTH, 2);
			this.setNumberPorts(dia.SOUTH, 1);
			this.vector1InPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'vector:',
				value : math.nullVector,
				defaultValue : math.nullVector,
				formatter : module.vectorFormatter,
				aggregator : module.vectorAggregator
			});
			this.vector2InPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'vector:',
				value : math.nullVector,
				defaultValue : math.nullVector,
				formatter : module.vectorFormatter,
				aggregator : module.vectorAggregator
			});
			this.vectorOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'vector:',
				value : math.nullVector,
				formatter : module.vectorFormatter,
				calculatePort : function() {
					var v1 = this.vector1InPort.value();
					var v2 = this.vector2InPort.value();
					var v = null;
					switch(this.properties.value.operator) {
						case '+':
							v = v1.add(v2);
							break;
						case '-':
							v = v1.subtract(v2);
							break;
					}
					return v;
				}
			});

		},

		text : function() {
			return this.properties.value.operator;
		},

		getProperties : function() {
			var self = this;
			return [this.getTitleProperty(), {
				label : 'operator',
				object : this.properties.value,
				name : 'operator',
				type : 'dropdown',
				values : ['+', '-']
			}];

		},

		endOfUpdate : function() {
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// OrbitalElements
	// =========================================================================

	module.OrbitalElements = dia.Element.extend({

		object : "OrbitalElements",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 200,
				title : 'Orbital Elements',
				value : {

					name : '-',

					semiMajorAxisOffset : 0,
					semiMajorAxisDelta : 0,

					eccentricityOffset : 0,
					eccentricityDelta : 0,

					inclinationOffset : 0,
					inclinationDelta : 0,

					meanLongitudeOffset : 0,
					meanLongitudeDelta : 0,

					longitudePerihelionOffset : 0,
					longitudePerihelionDelta : 0,

					longitudeAscendingNodeOffset : 0,
					longitudeAscendingNodeDelta : 0

				}
			}, properties);

			// ports
			this.setNumberPorts(dia.NORTH, 1);
			this.setNumberPorts(dia.EAST, 8);
			this.julianDateInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'Julian date:',
				value : 0,
				defaultValue : 0
			});
			this.semiMajorAxisOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'SemiMajorAxis:',
				value : 0,
				calculatePort : function() {
					return this.properties.value.semiMajorAxis;
				}
			});
			this.eccentricityOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Eccentricity:',
				value : 0,
				calculatePort : function() {
					return this.properties.value.eccentricity;
				}
			});
			this.inclinationOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Inclination:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.inclination;
				}
			});
			this.meanLongitudeOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Mean Longitude:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.meanLongitude;
				}
			});
			this.longitudePerihelionOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Longitude perihelion:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.longitudePerihelion;
				}
			});
			this.longitudeAscendingNodeOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Longitude ascending node:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.longitudeAscendingNode;
				}
			});
			this.meanAnomalyOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Mean anomaly:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.meanAnomaly;
				}
			});
			this.argumentPerihelionOutPort = this.createPort({
				location : dia.EAST,
				type : dia.OUT,
				label : 'Argument perihelion:',
				formatter : module.angleFormatter,
				value : 0,
				calculatePort : function() {
					return this.properties.value.argumentPerihelion;
				}
			});

		},

		text : function() {
			return this.properties.value.name;
		},

		getPlanets : function() {
			var planets = [];
			for (var planet in data) {
				planets.push(planet);
			}
			planets.push("-");
			return planets;
		},

		getPlanet : function() {
			return this.properties.value.name;
		},

		setPlanet : function(value) {
			this.properties.value.name = value;
			this.properties.value.semiMajorAxisOffset = data[value].semiMajorAxis[0];
			this.properties.value.semiMajorAxisDelta = data[value].semiMajorAxis[1];

			this.properties.value.eccentricityOffset = data[value].eccentricity[0];
			this.properties.value.eccentricityDelta = data[value].eccentricity[1];

			this.properties.value.inclinationOffset = data[value].inclination[0];
			this.properties.value.inclinationDelta = data[value].inclination[1];

			this.properties.value.meanLongitudeOffset = data[value].meanLongitude[0];
			this.properties.value.meanLongitudeDelta = data[value].meanLongitude[1];

			this.properties.value.longitudePerihelionOffset = data[value].longitudePerihelion[0];
			this.properties.value.longitudePerihelionDelta = data[value].longitudePerihelion[1];

			this.properties.value.longitudeAscendingNodeOffset = data[value].longitudeAscendingNode[0];
			this.properties.value.longitudeAscendingNodeDelta = data[value].longitudeAscendingNode[1];

			this.properties.title = value;

			this.refreshPropertyEditor();
			this.endOfUpdate();
		},

		getProperties : function() {
			var self = this;
			return [this.getTitleProperty(), {
				name : 'planet',
				label : 'planet',
				get : this.getPlanet,
				set : this.setPlanet,
				type : 'dropdown',
				values : this.getPlanets(),
				immediateInput : true
			}, {
				name : 'name',
				object : this.properties.value,
				label : 'name'
			}, {
				name : 'semiMajorAxisOffset',
				object : this.properties.value,
				label : 'semiMajorAxisOffset',
				convertFromUI : parseFloat
			}, {
				name : 'semiMajorAxisDelta',
				object : this.properties.value,
				label : 'semiMajorAxisDelta',
				convertFromUI : parseFloat
			}, {
				name : 'eccentricityOffset',
				object : this.properties.value,
				label : 'eccentricityOffset',
				convertFromUI : parseFloat
			}, {
				name : 'eccentricityDelta',
				object : this.properties.value,
				label : 'eccentricityDelta',
				convertFromUI : parseFloat
			}, {
				name : 'inclinationOffset',
				object : this.properties.value,
				label : 'inclinationOffset',
				convertFromUI : parseFloat
			}, {
				name : 'inclinationDelta',
				object : this.properties.value,
				label : 'inclinationDelta',
				convertFromUI : parseFloat
			}, {
				name : 'meanLongitudeOffset',
				object : this.properties.value,
				label : 'meanLongitudeOffset',
				convertFromUI : parseFloat
			}, {
				name : 'meanLongitudeDelta',
				object : this.properties.value,
				label : 'meanLongitudeDelta',
				convertFromUI : parseFloat
			}, {
				name : 'longitudePerihelionOffset',
				object : this.properties.value,
				label : 'longitudePerihelionOffset',
				convertFromUI : parseFloat
			}, {
				name : 'longitudePerihelionDelta',
				object : this.properties.value,
				label : 'longitudePerihelionDelta',
				convertFromUI : parseFloat
			}, {
				name : 'longitudeAscendingNodeOffset',
				object : this.properties.value,
				label : 'longitudeAscendingNodeOffset',
				convertFromUI : parseFloat
			}, {
				name : 'longitudeAscendingNodeDelta',
				object : this.properties.value,
				label : 'longitudeAscendingNodeDelta',
				convertFromUI : parseFloat
			}];

		},

		calculateOE : function(name, T) {
			return this.properties.value[name + 'Offset'] + T * this.properties.value[name + 'Delta'];
		},

		calculate : function() {
			var T = (this.julianDateInPort.value() - 2451545) / 36525.0;
			this.properties.value.semiMajorAxis = this.calculateOE('semiMajorAxis', T);
			this.properties.value.eccentricity = this.calculateOE('eccentricity', T);
			this.properties.value.inclination = math.deg2Rad(this.calculateOE('inclination', T));
			this.properties.value.meanLongitude = math.deg2Rad(this.calculateOE('meanLongitude', T));
			this.properties.value.longitudePerihelion = math.deg2Rad(this.calculateOE('longitudePerihelion', T));
			this.properties.value.longitudeAscendingNode = math.deg2Rad(this.calculateOE('longitudeAscendingNode', T));
			this.properties.value.meanAnomaly = this.properties.value.meanLongitude - this.properties.value.longitudePerihelion;
			this.properties.value.argumentPerihelion = this.properties.value.longitudePerihelion - this.properties.value.longitudeAscendingNode;
		},

		endOfUpdate : function() {
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Kepler Equation
	// =========================================================================

	module.Kepler = dia.Element.extend({

		object : "Kepler",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Kepler',
				value : math.nullVector
			}, properties);

			// ports
			this.setNumberPorts(dia.WEST, 3);
			this.setNumberPorts(dia.SOUTH, 1);
			this.semiMajorAxisInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'Semi major axis:',
				value : 0,
				defaultValue : 0
			});
			this.eccentricityInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'Eccentricity:',
				value : 0,
				defaultValue : 0
			});
			this.meanAnomalyInPort = this.createPort({
				location : dia.WEST,
				type : dia.IN,
				label : 'Mean anomaly:',
				value : 0,
				defaultValue : 0,
				formatter : module.angleFormatter
			});
			this.vectorOutPort = this.createPort({
				location : dia.SOUTH,
				type : dia.OUT,
				label : 'vector:',
				formatter : module.vectorFormatter,
				value : math.nullVector,
				calculatePort : function() {
					return this.properties.value;
				}
			});

		},

		text : function() {
			return "Kepler";
		},

		getProperties : function() {
			return [this.getTitleProperty()];
		},

		calculate : function() {
			var eccentricAnomaly = calc.eccentricAnomaly(this.meanAnomalyInPort.value(), this.eccentricityInPort.value());
			var lambda = 2 * Math.atan(Math.sqrt((1 + this.eccentricityInPort.value()) / (1 - this.eccentricityInPort.value())) * Math.tan(eccentricAnomaly / 2));
			var rad = this.semiMajorAxisInPort.value() * (1 - this.eccentricityInPort.value() * Math.cos(eccentricAnomaly));
			var v = math.vector(0, lambda, rad, 'spherical');
			this.properties.value.x = v.x;
			this.properties.value.y = v.y;
			this.properties.value.z = v.z;
		},

		endOfUpdate : function() {
			dia.Element.prototype.endOfUpdate.apply(this, arguments);
		}
	});

	// =========================================================================
	// Map
	// =========================================================================

	module.Map = dia.Element.extend({

		object : "Map",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Math',
				value : math.nullVector
			}, properties);

			// ports
			this.setNumberPorts(dia.NORTH, 1);
			this.vectorInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'Vector:',
				value : math.nullVector,
				defaultValue : math.nullVector,
				formatter : module.vectorFormatter,
				aggregator : module.vectorAggregator
			});

		},

		text : function() {
			return "Map";
		},

		getProperties : function() {
			return [this.getTitleProperty()];
		},

		calculate : function() {
		},

		showPropertyEditor : function() {
			$("#propertyeditor").children().remove();
			var paper = global.Raphael("propertyeditor", 1000, 1000);
			var map = global.astro.map.starMap({
				paper : paper
			});
			var value = this.vectorInPort.value();
			if ( value instanceof math.vector) {
				map.addPlanet(value);
			} else if ( value instanceof math.vectorList) {
				for (var i = 0; i < value.vectors.length; ++i) {
					map.addPlanet(value.vectors[i]);
				}
			}
            map.draw();
		}
	});

	// =========================================================================
	// Map2
	// =========================================================================

	module.Map2 = dia.Element.extend({

		object : "Map2",
		module : "astro.eph",

		init : function(properties) {

			// base init
			this.baseInit({
				width : 120,
				height : 60,
				title : 'Math',
				value : math.nullVector
			}, properties);

			// ports
			this.setNumberPorts(dia.NORTH, 1);
			this.vectorInPort = this.createPort({
				location : dia.NORTH,
				type : dia.IN,
				label : 'Vector:',
				value : math.nullVector,
				defaultValue : math.nullVector,
				formatter : module.vectorFormatter,
				aggregator : module.vectorAggregator
			});

		},

		text : function() {
			return "Map2";
		},

		getProperties : function() {
			return [this.getTitleProperty()];
		},

		calculate : function() {
		},

		showPropertyEditor : function() {
			$("#propertyeditor").children().remove();
			var paper = global.Raphael("propertyeditor", 1000, 1000);
			var map = global.astro.map.map({
				paper : paper,
				minVisibleLatitude : math.deg2Rad(-10)
			});
			map.drawBackground();			
			map.drawHourScale();
			map.drawEquator();

			var value = this.vectorInPort.value();
			if ( value instanceof math.vector) {
				map.addPlanet(value);
			} else if ( value instanceof math.vectorList) {
				for (var i = 0; i < value.vectors.length; ++i) {
					map.addPlanet(value.vectors[i]);
				}
			}
            map.draw();
		}
	});

})(this);

