/* global console, define, window, $, document */
define(["module",
	"./astro.core",
	"./astro.data",
	"./astro.math",
	"./astro.calc",
	"./constellations",
	"jquery",
	"raphael"], function (
		thisModule,
		core,
		data,
		math,
		calc,
		constellations,
		$,
		raphael) {

		"use strict";

		// =========================================================================
		// exported packages
		// =========================================================================
		var module = {};
		module.pathPrefix = thisModule.uri.slice(0, thisModule.uri.lastIndexOf("/") + 1);

		// =========================================================================
		// init star list
		// =========================================================================
		module.stars = $.map(constellations, function (star) {
			var starVector = math.vector(math.deg2Rad(star.latitude), math.deg2Rad(star.longitude), 1, "spherical");
			starVector.constellation = star.constellation;
			starVector.number = star.number;
			starVector.magnitude = star.magnitude;
			starVector.command = star.command;
			return starVector;
		});

		// =========================================================================
		// module load stars (deprecated)
		// =========================================================================
		module.loadStars = function (callback) {
			if (module.stars) {
				callback();
				return;
			}

			$.ajax({
				url: module.pathPrefix + "constellations.dat",
				success: function (data) {
					module.parseConstellations(data);
					callback();
				},
				dataType: 'text'
			});
		};

		// =========================================================================
		// parse constellations
		// =========================================================================
		module.parseConstellations = function (data) {
			module.stars = [];
			var records = data.split("\n");
			for (var i = 0; i < records.length; ++i) {

				// parse
				var parts = records[i].split(",");
				var star = {
					constellation: parts[0],
					number: parseInt(parts[1], 10),
					command: parseInt(parts[2], 10),
					longitude: parseFloat(parts[3]),
					latitude: parseFloat(parts[4]),
					magnitude: parseFloat(parts[5])
				};

				// convert to vector
				var starVector = math.vector(math.deg2Rad(star.latitude), math.deg2Rad(star.longitude), 1, "spherical");
				math.debug = false;

				starVector.constellation = star.constellation;
				starVector.number = star.number;
				starVector.magnitude = star.magnitude;
				starVector.command = star.command;
				starVector.latitude = star.latitude;
				starVector.longitude = star.longitude;

				// collect
				module.stars.push(starVector);

			}
		};

		// =========================================================================
		// vector to screen coordinate
		// =========================================================================
		module.screenCoordinatesCalculator = core.createClass({

			init: function (width, height, minLatitude, deltaLongitude, transformation) {
				this.mx = Math.round(width / 2);
				this.my = Math.round(height / 2);
				this.transformation = transformation;
				this.deltaLongitude = 0;
				this.minLatitude = minLatitude;
				if (deltaLongitude) {
					this.deltaLongitude = deltaLongitude;
				}
				var R_minLatitude = Math.round(Math.min(this.mx, this.my) * 1.0);
				this.R = Math.round(R_minLatitude / (1 - 2 * minLatitude / Math.PI));
			},

			setTransformation: function (transformation) {
				this.transformation = transformation;
			},

			calculate: function (object) {
				var transObject;
				if (this.transformation) {
					transObject = this.transformation(object);
				} else {
					transObject = object;
				}
				if (transObject.getLatitude() < this.minLatitude) {
					return null;
				}
				var r = this.R * (1 - 2 * transObject.getLatitude() / Math.PI);
				var x = Math.round(this.mx - r * Math.sin(transObject.getLongitude() - this.deltaLongitude));
				var y = Math.round(this.my + r * Math.cos(transObject.getLongitude() - this.deltaLongitude));
				return {
					vector: transObject,
					x: x,
					y: y
				};
			},

			calculateRotAngle: function (longitude) {
				return math.rad2Deg(longitude - this.deltaLongitude);
			},

			calculateRad: function (latitude) {
				return Math.round(this.R * (1 - 2 * latitude / Math.PI));
			},

			getCenter: function () {
				return {
					x: this.mx,
					y: this.my
				};
			}
		});

		// =========================================================================
		// magnitude to radius converter
		// =========================================================================
		module.magnitude2RadiusConverter = core.createClass({

			init: function (minRadius, maxRadius, minMagnitude, maxMagnitude) {
				this.maxRadius = maxRadius;
				this.minRadius = minRadius;
				this.maxMagnitude = maxMagnitude;
				this.minMagnitude = minMagnitude;
				this.m = (this.maxRadius - this.minRadius) / (this.maxMagnitude - this.minMagnitude);
			},

			calculate: function (magnitude) {
				if (magnitude <= this.minMagnitude) {
					return this.maxRadius;
				}
				if (magnitude >= this.maxMagnitude) {
					return this.minRadius;
				}
				return this.maxRadius - (magnitude - this.minMagnitude) * this.m;
			}
		});


		// =========================================================================
		// star map
		// =========================================================================
		module.starMap = core.createClass({

			init: function (opt) {

				this.daysVernalEquinox = 31 + 28 + 22 - 1; //+time/24 
				this.locationLatitude = math.deg2Rad(52); // Berlin
				this.minVisibleLatitude = -(Math.PI / 2 - this.locationLatitude);

				this.colorGray = "#6B6B6B";
				this.colorBlue = "#284dc7";
				this.colorRed = "#F70511";
				this.colorYellow = "#F4FF21";
				this.colorBlack = "#000000";
				this.colorWhite = "#FFFFFF";

				this.backGroundFill = null;
				this.backGroundStroke = this.colorGray;
				this.backGroundStrokeWidth = 0;

				this.hourScaleStroke = this.colorBlack;
				this.hourScaleStrokeWidth = 1;
				this.hourScaleFontSize = 15;
				this.hourScaleFontFamily = "Courier";
				this.hourScaleFontFill = this.colorBlack;
				this.hourScaleTickStrokeWidth = 2;
				this.hourScaleTickStroke = this.colorBlack;
				this.hourScaleLatitudeDelta = 25;

				this.equatorStroke = this.colorBlack;
				this.equatorStrokeWidth = 2;

				this.planetSize = 5;
				this.planetFill = this.colorRed;
				this.planetFontSize = 15;
				this.planetFontFamily = "Arial";

				this.paper = opt.paper;

				this.deltaLongitude = 0;

				this.planets = [];

				if (opt.day) {
					this.month = opt.month;
					this.day = opt.day;
					this.hour = opt.hour;
					this.minute = opt.minute;
				} else {
					this.month = 1;
					this.day = 1;
					this.hour = 0;
					this.minute = 0;
				}

				this.monthScaleFontSize = 15;
				this.monthScaleFontFamily = "Arial";
				this.monthScaleFontFill = this.colorBlack;
				this.monthScaleStroke = this.colorBlack;
				this.monthScaleStrokeWidth = 2;
				this.monthScaleTickStrokeWidth = 1;
				this.monthScaleTickStroke = this.colorBlack;
				this.monthScaleLatitudeDelta = 15;

				this.horizonStroke = this.colorBlue;
				this.horizonStrokeWidth = 2;

				this.horizonScaleStroke = this.colorBlue;
				this.horizonScaleStrokeWidth = 2;
				this.horizonScaleFontSize = 12;
				this.horizonScaleFontFamily = "Arial";
				this.horizonScaleFontFill = this.colorBlue;
				this.horizonScaleTickStrokeWidth = 1;
				this.horizonScaleTickStroke = this.colorBlue;
				this.horizonScaleLatitudeDelta = 12;

				this.eclipticStroke = this.colorRed;
				this.eclipticStrokeWidth = 1;
				this.eclipticScaleStroke = this.colorRed;
				this.eclipticScaleStrokeWidth = 1;

				this.eclipticScaleFontSize = 15;
				this.eclipticScaleFontFamily = "Arial";
				this.eclipticScaleFontFill = this.colorRed;
				this.eclipticScaleFontWeight = "normal";

				this.starLineStrokeWidth = 2;
				this.starLineStroke = this.colorGray;

				this.starFill = this.colorBlack;
				this.starStrokeWidth = 0;

				this.latitudeScaleStroke = this.colorBlue;
				this.latitudeScaleStrokeWidth = 2;

				this.latitudeScaleTickStroke = this.colorBlue;
				this.latitudeScaleTickStrokeWidth = 2;

				this.latitudeScaleFontSize = 15;
				this.latitudeScaleFontFamily = "Arial";
				this.latitudeScaleFontFill = this.colorBlue;
				this.latitudeScaleFontWeight = "bold";

			},

			drawBackground: function () {
				var center = this.screenCoordinatesCalculator.getCenter();
				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(this.minVisibleLatitude));
				if (this.backGroundFill) {
					circle.attr("fill", this.backGroundFill);
				}
				circle.attr("stroke", this.backGroundStroke);
				circle.attr("stroke-width", this.backGroundStrokeWidth);
			},

			drawHourScale: function () {

				var center = this.screenCoordinatesCalculator.getCenter();
				var deltaLongitude = 2 * Math.PI / (24 * 60);
				var totalMinutes = 0;
				var length = 0;
				var startLatitude = this.minVisibleLatitude - math.deg2Rad(this.hourScaleLatitudeDelta);

				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(startLatitude));
				circle.attr("fill-opacity", 0);
				circle.attr("stroke", this.hourScaleStroke);
				circle.attr("stroke-width", this.hourScaleStrokeWidth);

				for (var hour = 0; hour < 24; hour++) {
					for (var minutes = 0; minutes < 60; minutes += 5) {
						length = 2;
						var longitude = deltaLongitude * totalMinutes;
						if (minutes % 10 === 0) {
							length = 3;
						}
						if (minutes === 30) {
							length = 4;
						}
						if (minutes === 0) {
							length = 6;
							var pos = this.screenCoordinatesCalculator.calculateRad(startLatitude - math.deg2Rad(10));
							var text = this.paper.text(center.x, center.y, "" + hour);
							text.attr("font-size", this.hourScaleFontSize).attr("font-family", this.hourScaleFontFamily).attr("fill", this.hourScaleFontFill);
							text.transform("T0," + pos + "R" + this.screenCoordinatesCalculator.calculateRotAngle(longitude) + "," + center.x + "," + center.y);
							//text.translate(0, pos).rotate(this.screenCoordinatesCalculator.calculateRotAngle(longitude), center.x, center.y);
						}
						var line = this.drawLine(startLatitude, longitude, startLatitude - math.deg2Rad(length), longitude);
						line.attr("stroke-width", this.hourScaleTickStrokeWidth).attr("stroke", this.hourScaleTickStroke);
						totalMinutes += 5;
					}
				}
			},

			drawEquator: function () {
				var center = this.screenCoordinatesCalculator.getCenter();
				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(0));
				circle.attr("fill-opacity", 0);
				circle.attr("stroke", this.equatorStroke);
				circle.attr("stroke-width", this.equatorStrokeWidth);
			},

			drawLine: function (latitude1, longitude1, latitude2, longitude2) {
				var o1 = this.screenCoordinatesCalculator.calculate(math.vector(latitude1, longitude1, 1, "spherical"));
				var o2 = this.screenCoordinatesCalculator.calculate(math.vector(latitude2, longitude2, 1, "spherical"));
				return this.paper.path("M" + o1.x + " " + o1.y + "L" + o2.x + " " + o2.y);
			},

			addPlanet: function (planet) {
				this.planets.push(planet);
			},

			drawPlanets: function () {
				var self = this;
				var generatePlanetClickHandler = function (planet) {
					return function () {
						self.click(planet);
					};
				};
				for (var i = 0; i < this.planets.length; ++i) {
					var planet = this.planets[i];
					if (planet.getLatitude() < this.minVisibleLatitude) {
						return;
					}
					var pos = this.screenCoordinatesCalculator.calculate(planet);

					if (planet.title) {
						var text = this.paper.text(pos.x + 1.5 * this.planetSize, pos.y - 1.5 * this.planetSize, planet.title);
						text.attr("font-size", this.planetFontSize).attr("font-family", this.planetFontFamily).attr("fill", this.planetFill);
					}

					var circle = this.paper.circle(pos.x, pos.y, this.planetSize);
					circle.attr("fill", this.planetFill).attr("stroke-width", 0);
					circle.attr("title", planet.title + ' ' + planet.format({ type: 'spherical-hours', distance: false }))
					circle.click(generatePlanetClickHandler(planet));

				}
			},

			drawInit: function () {

				this.deltaLongitude = math.deg2Rad((calc.daysUpToMonth(this.month) + this.day - 1 - this.daysVernalEquinox) / 365.25 * 360);
				this.deltaLongitude += math.deg2Rad((this.hour + this.minute / 60 - 12) / 24 * 360);

				this.screenCoordinatesCalculator = module.screenCoordinatesCalculator(this.paper.width,
					this.paper.height,
					this.minVisibleLatitude - math.deg2Rad(40),
					this.deltaLongitude);

				this.paper.clear();

			},


			draw: function () {

				this.drawInit();

				this.drawBackground();
				this.drawMonthScale();
				this.drawHourScale();

				this.drawHorizon();
				this.drawHorizonScale();

				this.drawEquator();
				this.drawEcliptic();
				this.drawEclipticScale();
				this.drawStarLines();
				this.drawStars();
				this.drawPlanets();

			},

			drawMonthScale: function () {
				var center = this.screenCoordinatesCalculator.getCenter();
				var deltaLongitude = 2 * Math.PI / 365.25;
				var totalDays = 0;
				var length = 0;
				var startLatitude = this.minVisibleLatitude - math.deg2Rad(this.monthScaleLatitudeDelta);
				var offset = -this.daysVernalEquinox * deltaLongitude;

				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(startLatitude));
				circle.attr("fill-opacity", 0);
				circle.attr("stroke", this.monthScaleStroke);
				circle.attr("stroke-width", this.monthScaleStrokeWidth);

				for (var month = 1; month <= 12; month++) {
					var o3 = this.screenCoordinatesCalculator.calculateRad(startLatitude - math.deg2Rad(5));
					var text = this.paper.text(center.x, center.y, calc.nameOfMonth(month));
					text.attr("font-size", this.monthScaleFontSize).attr("font-family", this.monthScaleFontFamily).attr("fill", this.monthScaleFontFill);
					var longitude = (totalDays + 10) * deltaLongitude + offset;
					text.transform("T0," + o3 + "R" + this.screenCoordinatesCalculator.calculateRotAngle(longitude) + "," + center.x + "," + center.y);
					//text.translate(0, o3).rotate(this.screenCoordinatesCalculator.calculateRotAngle(longitude), center.x, center.y);
					for (var day = 1; day <= calc.daysOfMonth(month); day++) {
						longitude = totalDays * deltaLongitude + offset;
						length = 2;
						if (day % 5 === 0) {
							length = 4;
						}
						if (day === 1) {
							length = 8;
						}
						var line = this.drawLine(startLatitude, longitude, startLatitude - math.deg2Rad(length), longitude);
						line.attr("stroke-width", this.monthScaleTickStrokeWidth).attr("stroke", this.monthScaleTickStroke);
						/* if(month==3 && day==22){
							alert(totalDays+":"+longitude);
							var line = this.drawLine(3.14/2, longitude,-1,longitude);
							line.attr("stroke-width", this.monthScaleTickStrokeWidth).attr("stroke", this.monthScaleTickStroke);                        
						}*/
						totalDays += 1;
					}
				}
			},

			drawHorizon: function () {
				var oldV = null;
				var offset = this.deltaLongitude; // do not rotate horizon
				var center = this.screenCoordinatesCalculator.getCenter();
				var v;

				// horizon
				for (var longitude = 0; longitude < 360; longitude++) {
					v = math.vector(0, math.deg2Rad(longitude), 1, "spherical").rotX(-(Math.PI / 2 - this.locationLatitude)).rotZ(math.deg2Rad(-90) + offset);
					if (oldV !== null) {
						var p1 = this.screenCoordinatesCalculator.calculate(oldV);
						var p2 = this.screenCoordinatesCalculator.calculate(v);
						this.paper.path("M" + p1.x + " " + p1.y + "L" + p2.x + " " + p2.y).attr("stroke-width", this.horizonStrokeWidth).attr("stroke", this.horizonStroke);
					}
					oldV = v;
				}

				// directions N,S,W,E
				var directions = [{
					label: 'WEST',
					longitude: 0
				}, {
					label: 'SOUTH',
					longitude: 90
				}, {
					label: 'EAST',
					longitude: 180
				}, {
					label: 'NORTH',
					longitude: 270
				}];

				for (var i = 0; i < directions.length; ++i) {
					var direction = directions[i];
					v = math.vector(math.deg2Rad(-5), math.deg2Rad(direction.longitude), 1, "spherical").rotX(-(Math.PI / 2 - this.locationLatitude)).rotZ(math.deg2Rad(-90) + offset);
					var p = this.screenCoordinatesCalculator.calculate(v);
					var text = this.paper.text(p.x, p.y, "" + direction.label);
					text.attr("font-size", this.horizonScaleFontSize * 1.5).attr("font-family", this.horizonScaleFontFamily).attr("fill", this.horizonScaleFontFill);
					text.transform("R" + (direction.longitude - 90) + "," + p.x + "," + p.y);
					//text.rotate(direction.longitude - 90, p.x, p.y);
				}

			},

			drawHorizonScale: function () {
				var offset = math.deg2Rad(180) + this.deltaLongitude; // do not rotate horizon
				var center = this.screenCoordinatesCalculator.getCenter();
				var deltaLongitude = 2 * Math.PI / (24 * 60);
				var totalMinutes = 0;
				var length = 0;
				var startLatitude = this.minVisibleLatitude - math.deg2Rad(this.horizonScaleLatitudeDelta);

				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(startLatitude));
				circle.attr("fill-opacity", 0);
				circle.attr("stroke", this.horizonScaleStroke);
				circle.attr("stroke-width", this.horizonScaleStrokeWidth);

				for (var hour = 0; hour < 24; hour++) {
					for (var minutes = 0; minutes < 60; minutes += 5) {
						var longitude = deltaLongitude * totalMinutes + offset;
						length = 2;
						if (minutes % 10 === 0) {
							length = 3;
						}
						if (minutes === 30) {
							length = 4;
						}
						if (minutes === 0) {
							length = 6;
							var pos = this.screenCoordinatesCalculator.calculateRad(startLatitude + math.deg2Rad(8));
							var displayHour = 24 - hour;
							if (displayHour === 24) {
								displayHour = 0;
							}
							var text = this.paper.text(center.x, center.y, "" + displayHour);
							text.attr("font-size", this.horizonScaleFontSize).attr("font-family", this.horizonScaleFontFamily).attr("fill", this.horizonScaleFontFill);
							text.transform("T0," + pos + "R" + this.screenCoordinatesCalculator.calculateRotAngle(longitude) + "," + center.x + "," + center.y);
							//text.translate(0, pos).rotate(this.screenCoordinatesCalculator.calculateRotAngle(longitude), center.x, center.y);
						}
						var line = this.drawLine(startLatitude, longitude, startLatitude + math.deg2Rad(length), longitude);
						line.attr("stroke-width", this.horizonScaleTickStrokeWidth).attr("stroke", this.horizonScaleTickStroke);
						totalMinutes += 5;
					}
				}

			},

			drawEcliptic: function () {
				var oldV = null;
				for (var lat = 0; lat < 360; lat++) {
					var v = math.vector(0, math.deg2Rad(lat), 1,
						"spherical").rotX(calc.eclipticAngle0);
					if (oldV !== null) {
						var p1 = this.screenCoordinatesCalculator.calculate(oldV);
						var p2 = this.screenCoordinatesCalculator.calculate(v);
						this.paper.path("M" + p1.x + " " + p1.y + "L" + p2.x + " " + p2.y).attr("stroke-width", this.eclipticStrokeWidth).attr("stroke", this.eclipticStroke);
					}
					oldV = v;
				}
			},

			drawEclipticScale: function () {
				var length = 0;
				var center = this.screenCoordinatesCalculator.getCenter();
				for (var lat = 0; lat < 360; lat += 5) {
					length = 1;
					var v = math.vector(0, math.deg2Rad(lat), 1,
						"spherical").rotX(calc.eclipticAngle0);
					if (lat % 10 === 0) {
						length = 2;
					}
					if (lat % 50 === 0) {
						var pos = this.screenCoordinatesCalculator.calculateRad(v.getLatitude() - math.deg2Rad(5));
						var text = this.paper.text(center.x, center.y, "" + lat);
						text.attr("font-size", this.eclipticScaleFontSize).attr("font-family", this.eclipticScaleFontFamily).attr("fill", this.eclipticScaleFontFill).attr("font-weight", this.eclipticScaleFontWeight);
						text.transform("T0," + pos + "R" + this.screenCoordinatesCalculator.calculateRotAngle(v.getLongitude()) + "," + center.x + "," + center.y);
						//text.translate(0, pos).rotate(this.screenCoordinatesCalculator.calculateRotAngle(v.getLongitude()), center.x, center.y);
					}
					var v1 = math.vector(v.getLatitude() - math.deg2Rad(length), v.getLongitude(), 1, "spherical");
					var v2 = math.vector(v.getLatitude() + math.deg2Rad(length), v.getLongitude(), 1, "spherical");
					var p1 = this.screenCoordinatesCalculator.calculate(v1);
					var p2 = this.screenCoordinatesCalculator.calculate(v2);
					this.paper.path("M" + p1.x + " " + p1.y + "L" + p2.x + " " + p2.y).attr("stroke-width", this.eclipticScaleStrokeWidth).attr("stroke", this.eclipticScaleStroke);
				}
			},

			drawStars: function () {
				var self = this;
				var generateStarClickHandler = function (star) {
					return function () {
						self.click(star);
					};
				};
				var magnitude2RadiusConverter = module.magnitude2RadiusConverter(2, 8, -2, 5);
				for (var i = 0; i < module.stars.length; ++i) {
					var star = module.stars[i];
					if (star.getLatitude() < this.minVisibleLatitude) {
						continue;
					}
					var pos = this.screenCoordinatesCalculator.calculate(star);
					var rad = magnitude2RadiusConverter.calculate(star.magnitude);
					var circle = this.paper.circle(pos.x, pos.y, rad);
					circle.attr("fill", this.starFill).attr("stroke-width", this.starStrokeWidth);
					circle.attr("title", star.constellation + ' ' + star.format({ type: 'spherical-hours', distance: false }));
					circle.click(generateStarClickHandler(star));
				}
			},

			click: function (obj) {

				var center = this.screenCoordinatesCalculator.getCenter();
				var startLatitude = math.rad2Deg(this.minVisibleLatitude) - this.hourScaleLatitudeDelta;

				// line
				var line = this.drawLine(math.deg2Rad(startLatitude), obj.getLongitude(), Math.PI / 2, obj.getLongitude());
				line.attr("stroke-width", this.latitudeScaleStrokeWidth).attr("stroke", this.latitudeScaleStroke);

				// scale
				var deltaLatitude = 10;
				for (var latitude = 90; latitude >= startLatitude; latitude -= deltaLatitude) {

					// tick
					var pos = this.screenCoordinatesCalculator.calculateRad(math.deg2Rad(latitude));
					line = this.paper.path("M" + (center.x - 10) + " " + center.y + "L" + (center.x + 10) + " " + center.y);
					line.transform("T0," + pos + "R" + this.screenCoordinatesCalculator.calculateRotAngle(obj.getLongitude()) + "," + center.x + "," + center.y);
					//line.translate(0, pos).rotate(this.screenCoordinatesCalculator.calculateRotAngle(obj.getLongitude()), center.x, center.y);
					line.attr("stroke-width", this.latitudeScaleTickStrokeWidth).attr("stroke", this.latitudeScaleTickStroke);

					// label of tick
					var text = this.paper.text(center.x + 20, center.y, "" + latitude);
					text.transform("T0," + pos + "R" + this.screenCoordinatesCalculator.calculateRotAngle(obj.getLongitude()) + "," + center.x + "," + center.y);
					//text.translate(0, pos).rotate(this.screenCoordinatesCalculator.calculateRotAngle(obj.getLongitude()), center.x, center.y);
					text.attr("font-size", this.latitudeScaleFontSize).attr("font-family", this.latitudeScaleFontFamily).attr("fill", this.latitudeScaleFontFill).attr("font-weight", this.latitudeScaleFontWeight);
				}
			},

			drawStarLines: function () {
				var star = null;
				var oldPosition = null;
				for (var i = 0; i < module.stars.length; ++i) {
					star = module.stars[i];
					switch (star.command) {
						case 0:
							oldPosition = star;
							break;
						case 1:
							if (star.getLatitude() > this.minVisibleLatitude && oldPosition.getLatitude() > this.minVisibleLatitude) {
								var pos1 = this.screenCoordinatesCalculator.calculate(oldPosition);
								var pos2 = this.screenCoordinatesCalculator.calculate(star);
								this.paper.path("M" + pos1.x + " " + pos1.y + "L" + pos2.x + " " + pos2.y).attr("stroke-width", this.starLineStrokeWidth).attr("stroke", this.starLineStroke);
							}
							oldPosition = star;
							break;
					}
				}
			}
		});

		// =========================================================================
		// polar map
		// =========================================================================
		module.polarMap = core.createClass({

			init: function (opt) {
				this.paper = opt.paper;
				this.opt = opt;
				opt.realWidth = opt.realWidth || 4;
				opt.realHeight = opt.realHeight || 4;
				opt.cx = this.paper.width / 2;
				opt.cy = this.paper.height / 2;
				opt.planetSize = this.paper.width / 50;
			},

			calculateX: function (x, withOffset) {
				if (withOffset) {
					return this.opt.cx + x / this.opt.realWidth * this.paper.width;
				} else {
					return x / this.opt.realWidth * this.paper.width;
				}
			},

			calculateY: function (y, withOffset) {
				if (withOffset) {
					return this.opt.cy - y / this.opt.realHeight * this.paper.height;
				} else {
					return y / this.opt.realHeight * this.paper.height;
				}
			},

			drawPlanet: function (v) {
				// orbit
				var r = Math.sqrt(v.x * v.x + v.y * v.y);
				this.paper.ellipse(this.opt.cx, this.opt.cy, this.calculateX(r, false), this.calculateY(r, false));
				// planet
				var x = this.calculateX(v.x, true);
				var y = this.calculateY(v.y, true);
				var planet = this.paper.circle(x, y, this.opt.planetSize);
				planet.attr("fill", "black");
				// label
				if (v.title) {
					this.paper.text(x + 1.5 * this.opt.planetSize, y - 1.5 * this.opt.planetSize, v.title);
				}
			},

			drawScale: function (v, rad) {
				rad = rad || 0.5;
				var tickLength = rad / 20;
				// circle for ticks
				var x = this.calculateX(v.x, true);
				var y = this.calculateY(v.y, true);
				this.paper.ellipse(x, y, this.calculateX(rad, false), this.calculateY(rad, false));
				// ticks
				var p1, p2;
				for (var a = 0; a < 360; a += 5) {
					if (a % 10 === 0) {
						p1 = v.add(math.vector(0, math.deg2Rad(a), rad + tickLength, "spherical"));
						p2 = v.add(math.vector(0, math.deg2Rad(a), rad - tickLength, "spherical"));
						this.drawLine(p1, p2);
					} else if (a % 5 === 0) {
						p1 = v.add(math.vector(0, math.deg2Rad(a), rad + tickLength / 2, "spherical"));
						p2 = v.add(math.vector(0, math.deg2Rad(a), rad - tickLength / 2, "spherical"));
						this.drawLine(p1, p2);
					}
					if (a % 20 === 0) {
						var text = this.paper.text(x, y, "" + a);
						text.transform("T0," + this.calculateY(rad + 2 * tickLength, false) + "R" + (-90 - a) + "," + x + "," + y);
						//text.translate(0, trans);                    
						//text.rotate(-90 - a, x, y);
					}
				}
			},

			drawLine: function (v1, v2) {
				var pos1 = {
					x: this.calculateX(v1.x, true),
					y: this.calculateY(v1.y, true),
				};
				var pos2 = {
					x: this.calculateX(v2.x, true),
					y: this.calculateY(v2.y, true),
				};
				return this.paper.path("M" + pos1.x + " " + pos1.y + "L" + pos2.x + " " + pos2.y);
			}
		});

		// =========================================================================
		// interactive map
		// =========================================================================
		module.interactiveMap = core.createClass({

			init: function (opt) {

				this.opt = opt;
				var self = this;

				// map
				var mapContainer = $("<div></div>");
				this.opt.parent.append(mapContainer);

				this.paper = raphael(mapContainer[0], mapContainer.width(), mapContainer.width());
				this.map = module.starMap({
					paper: this.paper,
					day: opt.day,
					month: opt.month,
					hour: opt.hour,
					minute: opt.minute
				});

				// toolbar
				var toolbar = $("<div></div>");
				this.opt.parent.prepend(toolbar);
				var table = $("<table></table>");
				toolbar.append(table);

				// helper input field creation
				var createInputField = function (table, label, value) {
					var row = $("<tr></tr>");
					table.append(row);
					var cell = $("<td></td>");
					row.append(cell);
					cell.text(label);
					cell = $("<td></td>");
					row.append(cell);
					var input = $("<input type='text'></input>");
					input.val(value);
					cell.append(input);
					return input;
				};

				// helper checkbox creation
				var createCheckBox = function (table, label, value, callback) {
					var row = $("<tr></tr>");
					table.append(row);
					var cell = $("<td></td>");
					row.append(cell);
					cell.text(label);
					cell = $("<td></td>");
					row.append(cell);
					var checkBox = $("<input type='checkbox'></input>");
					checkBox.prop("checked", value);
					cell.append(checkBox);
					return checkBox;
				};

				// input fields
				this.day = createInputField(table, "Day", this.map.day);
				this.month = createInputField(table, "Month", this.map.month);
				this.hour = createInputField(table, "Hour", this.map.hour);
				this.minute = createInputField(table, "Minute", this.map.minute);

				// checkboxes
				this.flgDateScale = createCheckBox(table, "Date scale", true);
				this.flgHourScale = createCheckBox(table, "Rectascension", true);
				this.flgHorizon = createCheckBox(table, "Horizon", true);
				this.flgEquator = createCheckBox(table, "Equator", true);
				this.flgEcliptic = createCheckBox(table, "Ecliptic", true);
				this.flgStars = createCheckBox(table, "Stars", true);
				this.flgPlanets = createCheckBox(table, "Planets", true);

				// update button
				var updateButton = $("<button>Update</button>");
				toolbar.append(updateButton);
				updateButton.click(function () {
					self.draw();
				});

				// link
				var linkContainer = $("<div></div>");
				toolbar.append(linkContainer);
				this.link = $("<a href='' target='_blank'>Open in new Window</a>");
				linkContainer.append(this.link);

			},

			draw: function () {

				this.map.day = parseInt(this.day.val(), 10);
				this.map.month = parseInt(this.month.val(), 10);
				this.map.hour = parseInt(this.hour.val(), 10);
				this.map.minute = parseInt(this.minute.val(), 10);

				var url = module.pathPrefix + "../html/map.html?data=" + core.url.prototype.encode(JSON.stringify(this.toJson()));
				this.link.attr("href", url);

				this.map.drawInit();
				this.map.drawBackground();
				if (this.flgDateScale.is(":checked")) {
					this.map.drawMonthScale();
				}
				if (this.flgHourScale.is(":checked")) {
					this.map.drawHourScale();
				}
				if (this.flgHorizon.is(":checked")) {
					this.map.drawHorizon();
					this.map.drawHorizonScale();
				}
				if (this.flgEquator.is(":checked")) {
					this.map.drawEquator();
				}
				if (this.flgEcliptic.is(":checked")) {
					this.map.drawEcliptic();
					this.map.drawEclipticScale();
				}
				if (this.flgStars.is(":checked")) {
					this.map.drawStarLines();
					this.map.drawStars();
				}
				if (this.flgPlanets.is(":checked")) {
					this.map.drawPlanets();
				}

			},

			toJson: function () {

				var json = {};

				// planets
				var planets = [];
				for (var i = 0; i < this.map.planets.length; ++i) {
					var planet = this.map.planets[i];
					planets.push(planet.toJson());
				}
				json.planets = planets;

				// date
				json.day = this.map.day;
				json.month = this.map.month;
				json.hour = this.map.hour;
				json.minute = this.map.minute;

				return json;
			},

			fromJson: function (json) {
				// planets
				for (var i = 0; i < json.planets.length; ++i) {
					var planetJson = json.planets[i];
					var planet = math.vector();
					planet.fromJson(planetJson);
					this.map.addPlanet(planet);
				}
				// date
				this.map.day = json.day;
				this.map.month = json.month;
				this.map.hour = json.hour;
				this.map.minute = json.minute;
				this.day.val(this.map.day);
				this.month.val(this.map.month);
				this.minute.val(this.map.minute);
				this.hour.val(this.map.hour);
			},

			fromUrl: function () {
				var json = core.url().parameterJSON("data");
				if (!json) {
					return;
				}
				this.fromJson(json);
			},

			addPlanet: function (v) {
				this.map.addPlanet(v);
			}

		});


		// =========================================================================
		// simple map
		// =========================================================================
		module.simpleMap = core.createClass({

			init: function (opt) {

				this.colorGray = "#6B6B6B";
				this.colorBlue = "#284dc7";
				this.colorRed = "#F70511";
				this.colorYellow = "#F4FF21";
				this.colorBlack = "#000000";
				this.colorWhite = "#FFFFFF";

				this.paper = opt.paper;

				this.minVisibleLatitude = -math.deg2Rad(25);

				this.backGroundFill = null;
				this.backGroundStroke = this.colorGray;
				this.backGroundStrokeWidth = 1;

				this.equatorStroke = this.colorBlack;
				this.equatorStrokeWidth = 2;

				this.eclipticStroke = this.colorBlue;
				this.eclipticStrokeWidth = 1;

				this.starLineStrokeWidth = 2;
				this.starLineStroke = this.colorGray;

				this.starFill = this.colorBlack;
				this.starStrokeWidth = 0;

				this.planetSize = 5;
				this.planetFill = this.colorRed;
				this.planetFontSize = 15;
				this.planetFontFamily = "Arial";

				this.tranformation = opt.tranformation;

				this.planets = [];

				this.magnitude2RadiusConverter = module.magnitude2RadiusConverter(2, 8, -2, 5);

			},

			drawInit: function () {
				this.screenCoordinatesCalculator = module.screenCoordinatesCalculator(this.paper.width - 10,
					this.paper.height - 10,
					this.minVisibleLatitude, 0, this.tranformation);
				this.paper.clear();
			},

			setTransformation: function (transformation) {
				this.transformation = transformation;
				this.screenCoordinatesCalculator.setTransformation(transformation);
			},

			setGeoLocation: function (siderealTime, geoLatitude) {
				this.setTransformation(function (obj) {
					return obj.rotZ(-siderealTime).rotY(-Math.PI / 2 + geoLatitude);
				});
			},

			drawBackground: function () {
				var center = this.screenCoordinatesCalculator.getCenter();
				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(this.minVisibleLatitude));
				if (this.backGroundFill) {
					circle.attr("fill", this.backGroundFill);
				}
				circle.attr("stroke", this.backGroundStroke);
				circle.attr("stroke-width", this.backGroundStrokeWidth);
			},

			drawEquator: function () {
				var center = this.screenCoordinatesCalculator.getCenter();
				var circle = this.paper.circle(center.x, center.y, this.screenCoordinatesCalculator.calculateRad(0));
				circle.attr("fill-opacity", 0);
				circle.attr("stroke", this.equatorStroke);
				circle.attr("stroke-width", this.equatorStrokeWidth);
			},

			drawLine: function (latitude1, longitude1, latitude2, longitude2) {
				var o1, o2;
				if (latitude1 instanceof math.vector) {
					o1 = this.screenCoordinatesCalculator.calculate(latitude1);
					o2 = this.screenCoordinatesCalculator.calculate(longitude1);
				} else {
					o1 = this.screenCoordinatesCalculator.calculate(math.vector(latitude1, longitude1, 1, "spherical"));
					o2 = this.screenCoordinatesCalculator.calculate(math.vector(latitude2, longitude2, 1, "spherical"));
				}
				if (o1 && o2) {
					return this.paper.path("M" + o1.x + " " + o1.y + "L" + o2.x + " " + o2.y);
				}
			},

			drawCircle: function (position, rad) {
				var o1 = this.screenCoordinatesCalculator.calculate(position);
				if (o1) {
					var circle = this.paper.circle(o1.x, o1.y, rad);
					return circle;
				}
			},

			drawText: function (position, text, deltaX, deltaY) {
				var o1 = this.screenCoordinatesCalculator.calculate(position);
				if (o1) {
					var text = this.paper.text(o1.x + deltaX, o1.y + deltaY, text);
					return text;
				}
			},

			addPlanet: function (planet) {
				this.planets.push(planet);
			},

			drawPlanet: function (planet) {

				if (planet.title) {
					var text = this.drawText(planet, planet.title, 0, -10);
					if (text) {
						if (planet.color) {
							text.attr("fill", planet.color);
						} else {
							text.attr("fill", this.planetFill);
						}
						text.attr('font-size', this.planetFontSize);
					}
				}

				var circle = this.drawCircle(planet, this.planetSize);
				if (circle) {
					if (planet.color) {
						circle.attr("fill", planet.color).attr("stroke-width", 0);
					} else {
						circle.attr("fill", this.planetFill).attr("stroke-width", 0);
					}
					var tooltip = (planet.title ? planet.title + ' ' : '') + planet.format({ type: 'spherical-hours', distance: false });
					circle.attr('title', tooltip);
				}

			},

			drawPlanets: function () {
				for (var i = 0; i < this.planets.length; ++i) {
					var planet = this.planets[i];
					this.drawPlanet(planet);
				}
			},

			drawEcliptic: function () {
				var oldV = null;
				for (var lat = 0; lat < 360; lat++) {
					var v = math.vector(0, math.deg2Rad(lat), 1,
						"spherical").rotX(calc.eclipticAngle0);
					if (oldV !== null) {
						var p1 = this.screenCoordinatesCalculator.calculate(oldV);
						var p2 = this.screenCoordinatesCalculator.calculate(v);
						if (p1 && p2) {
							this.paper.path("M" + p1.x + " " + p1.y + "L" + p2.x + " " + p2.y)
								.attr("stroke-width", this.eclipticStrokeWidth).attr("stroke", this.eclipticStroke);
						}

					}
					oldV = v;
				}
			},

			drawStar: function (star) {
				var rad = this.magnitude2RadiusConverter.calculate(star.magnitude);
				var circle = this.drawCircle(star, rad);
				if (!circle) {
					return;
				}
				circle.attr("fill", this.starFill).attr("stroke-width", this.starStrokeWidth);
				circle.attr("title", star.format({ type: 'spherical-hours', distance: false }))
			},

			drawStars: function () {
				var self = this;
				for (var i = 0; i < module.stars.length; ++i) {
					var star = module.stars[i];
					this.drawStar(star);
				}

			},

			drawStarLines: function () {
				var star = null;
				var oldPosition = null;
				for (var i = 0; i < module.stars.length; ++i) {
					star = module.stars[i];
					switch (star.command) {
						case 0:
							oldPosition = star;
							break;
						case 1:
							var pos1 = this.screenCoordinatesCalculator.calculate(oldPosition);
							var pos2 = this.screenCoordinatesCalculator.calculate(star);
							if (!pos1 || !pos2) {
								continue;
							}
							this.paper.path("M" + pos1.x + " " + pos1.y + "L" + pos2.x + " " + pos2.y)
								.attr("stroke-width", this.starLineStrokeWidth).attr("stroke", this.starLineStroke);
							oldPosition = star;
							break;
					}
				}
			},

			draw: function (what) {
				what = what || ['equator', 'ecliptic', 'stars', 'planets'];
				this.drawInit();
				this.drawBackground();
				if (what.indexOf('equator') >= 0) {
					this.drawEquator();
				}
				if (what.indexOf('ecliptic') >= 0) {
					this.drawEcliptic();
				}
				if (what.indexOf('stars') >= 0) {
					this.drawStarLines();
					this.drawStars();
				}
				if (what.indexOf('planets') >= 0) {
					this.drawPlanets();
				}
			}

		});

		return module;

	});