/* global console, define, window, $, document */
define(["./astro.core", "jquery"], function (core, $) {

    "use strict";

    // =========================================================================
    // packages
    // =========================================================================
    var math = {};

    // =========================================================================
    // conversion degree -> radian
    // =========================================================================

    math.deg2Rad = function (deg) {
        return deg * Math.PI / 180;
    };

    // =========================================================================
    // conversion radian -> degree
    // =========================================================================

    math.rad2Deg = function (rad) {
        return rad * 180 / Math.PI;
    };

    // =========================================================================
    // limit angle to 0..2pi
    // =========================================================================

    math.limit02Pi = function (rad) {
        if (rad < 0) {
            rad += (Math.floor(Math.abs(rad) / (2 * Math.PI)) + 1) * 2 * Math.PI;
        }
        var n = math.int(rad / (2 * Math.PI));
        return rad - n * 2 * Math.PI;
    };

    // =========================================================================
    // integer of number
    // =========================================================================

    math.int = function (n) {
        if (n > 0) {
            return Math.floor(n);
        } else {
            return -Math.floor(Math.abs(n));
        }
    };


    math.log10 = function (x) {
        return Math.log(x) / Math.LN10;
    };

    // =========================================================================
    // trim length of number
    // =========================================================================

    math.BLANKS1 = "                                ";
    math.BLANKS2 = "00000000000000000000000000000000";

    math.trimLength = function (length, text, char) {
        text = text.toString();
        var padding = length - text.length;
        if (padding <= 0) {
            return text;
        }
        if (char === undefined) {
            return math.BLANKS1.slice(0, padding) + text;
        } else {
            return math.BLANKS2.slice(0, padding) + text;
        }

    };

    // =========================================================================
    // format angle
    // =========================================================================

    math.formatAngle = function (format, angle) {
        var m, s, rest;
        if (format === 'degrees' || format === 'degrees-latitude') {
            rest = math.rad2Deg(math.limit02Pi(angle));
            if (format === 'degrees-latitude' && rest > 90) {
                rest -= 360;
            }
            var sign = null;
            if (rest >= 0) {
                sign = "+";
            } else {
                sign = "-";
            }
            rest = Math.abs(rest);
            var d = math.int(rest);
            rest = (rest - d) * 60;
            m = math.int(rest);
            rest = (rest - m) * 60;
            s = math.int(rest);
            return sign + math.trimLength(3, d) + ":" + math.trimLength(2, m, '0') + ":" + math.trimLength(2, s, '0');
        } else if (format === 'degrees-simple' || format === 'degrees-simple-latitude') {
            var a = math.rad2Deg(math.limit02Pi(angle));
            if (format === 'degrees-simple-latitude' && a > 90) {
                a -= 360;
            }
            s = math.trimLength(8, a.toFixed(2));
            return s;
        } else if (format === 'hours') {
            rest = null;
            rest = math.rad2Deg(math.limit02Pi(angle)) * 24 / 360.0;
            var h = math.int(rest);
            rest = (rest - h) * 60;
            m = math.int(rest);
            rest = (rest - m) * 60;
            s = math.int(rest);
            return "  " + math.trimLength(2, h) + ":" + math.trimLength(2, m, '0') + ":" + math.trimLength(2, s, '0');
        } else {
            throw "Unknown format '" + format + "' (use 'degrees' or 'hours')";
        }
    };

    // =========================================================================
    // parse angle
    // =========================================================================

    math.parseAngle = function (format, angleString) {
        angleString = $.trim(angleString);
        var sign = +1;
        if (angleString[0] === '+' || angleString[0] === '-') {
            switch (angleString[0]) {
                case '+':
                    sign = +1;
                    break;
                case '-':
                    sign = -1;
                    break;
            }
            angleString = angleString.slice(1);
        }
        switch (format) {
            case 'degrees':
                var splitted = angleString.split(":");
                var degrees = parseInt(splitted[0], 10);
                var minutes = parseInt(splitted[1], 10);
                var seconds = parseInt(splitted[2], 10);
                return sign * (degrees + minutes / 60.0 + seconds / 3600.0) * Math.PI / 180;
            case 'degrees-simple':
                var a = parseFloat(angleString);
                return sign * math.deg2Rad(a);
            case 'hours':
                splitted = angleString.split(":");
                var hours = parseInt(splitted[0], 10);
                minutes = parseInt(splitted[1], 10);
                seconds = parseInt(splitted[2], 10);
                return sign * (hours + minutes / 60.0 + seconds / 3600.0) * Math.PI / 12;
        }
    };

    // =========================================================================
    // vector
    // =========================================================================

    math.vector = core.createClass({

        init: function (x, y, z, mode) {
            if (arguments.length === 0) {
                this.x = this.y = this.z = 0;
            } else if (arguments.length === 3) {
                this.x = x;
                this.y = y;
                this.z = z;
            } else if (arguments.length === 4) {
                if (mode === 'cartesian') {
                    this.x = x;
                    this.y = y;
                    this.z = z;
                } else if (mode === 'spherical') {
                    // x=beta / y=lambda / z=r
                    this.x = z * Math.cos(x) * Math.cos(y);
                    this.y = z * Math.cos(x) * Math.sin(y);
                    this.z = z * Math.sin(x);
                } else {
                    throw "Illegal vector construction / mode unknown / use 'cartesian' or 'spherical'";
                }
            } else {
                throw "Illegal vector construction";
            }
        },

        toJson: function () {
            var json = {};
            for (var a in this) {
                if (this.hasOwnProperty(a)) {
                    json[a] = this[a];
                }
            }
            return json;
        },

        fromJson: function (json) {
            $.extend(this, json);
        },

        getLongitude: function () {

            if (this.x > 0 && this.y > 0)
                return Math.atan(this.y / this.x);
            if (this.x > 0 && this.y < 0)
                return 2 * Math.PI + Math.atan(this.y / this.x);
            if (this.x > 0 && this.y === 0)
                return 0;

            if (this.x === 0 && this.y > 0)
                return Math.PI / 2;
            if (this.x === 0 && this.y < 0)
                return 3 * Math.PI / 2;
            if (this.x === 0 && this.y === 0)
                return 0;

            if (this.x < 0 && this.y > 0)
                return Math.PI + Math.atan(this.y / this.x);
            if (this.x < 0 && this.y < 0)
                return Math.PI + Math.atan(this.y / this.x);
            if (this.x < 0 && this.y === 0)
                return Math.PI;

            return 0;
        },

        getLatitude: function () {
            var rho = Math.sqrt(this.x * this.x + this.y * this.y);
            if (rho !== 0) {
                return Math.atan(this.z / rho);
            } else {
                if (this.z < 0)
                    return -Math.PI / 2;
                if (this.z > 0)
                    return Math.PI / 2;
                if (this.z === 0)
                    return 0;
            }
            return 0;
        },

        getDistance: function () {
            return Math.sqrt(this.x * this.x + this.y * this.y + this.z * this.z);
        },

        getDistance2: function () {
            return this.x * this.x + this.y * this.y + this.z * this.z;
        },

        add: function (other) {
            return math.vector(this.x + other.x, this.y + other.y, this.z + other.z);
        },

        subtract: function (other) {
            return math.vector(this.x - other.x, this.y - other.y, this.z - other.z);
        },

        multiply: function (f) {
            return math.vector(f * this.x, f * this.y, f * this.z);
        },

        normalize: function () {
            var r = this.getDistance();
            return math.vector(this.x / r, this.y / r, this.z / r);
        },

        scalarProduct: function (other) {
            return this.x * other.x + this.y * other.y + this.z * other.z;
        },

        invert: function () {
            return math.vector(-this.x, -this.y, -this.z);
        },

        rotX: function (alpha) {
            var result = math.vector();
            result.x = this.x;
            result.y = Math.cos(alpha) * this.y - Math.sin(alpha) * this.z;
            result.z = Math.sin(alpha) * this.y + Math.cos(alpha) * this.z;
            return result;
        },

        rotY: function (alpha) {
            var result = math.vector();
            result.x = Math.cos(alpha) * this.x + Math.sin(alpha) * this.z;
            result.y = this.y;
            result.z = -Math.sin(alpha) * this.x + Math.cos(alpha) * this.z;
            return result;
        },

        rotZ: function (alpha) {
            var result = math.vector();
            result.x = Math.cos(alpha) * this.x - Math.sin(alpha) * this.y;
            result.y = Math.sin(alpha) * this.x + Math.cos(alpha) * this.y;
            result.z = this.z;
            return result;
        },

        mirrorX: function (alpha) {
            var result = math.vector();
            result.x = this.x;
            result.y = -this.y;
            result.z = this.z;
            return result;
        },

        mirrorY: function (alpha) {
            var result = math.vector();
            result.x = -this.x;
            result.y = this.y;
            result.z = this.z;
            return result;
        },

        angle: function (other) {
            var diff = this.subtract(other);
            return math.limit02Pi(Math.acos(this.scalarProduct(other) / this.getDistance() / other.getDistance()));
        },

        toString: function () {
            return "Vector";
        },

        format: function (input) {

            var options;
            if (typeof input === 'object') {
                options = input
                if (typeof options.distance === 'undefined') {
                    options.distance = true;
                }
            } else {
                options = {
                    type: input,
                    distance: true
                }
            }

            if (options.type === 'spherical-degrees') {
                return this.formatDegrees(options);
            } else if (options.type === 'spherical-degrees-simple') {
                return this.formatDegreesSimple(options);
            } else if (options.type === 'spherical-hours') {
                return this.formatHours(options);
            } else if (options.type === 'cartesian') {
                return this.formatCartesian(options);
            } else {
                throw "Unknown formatting: ''" + format.type + "' (use 'spherical-degrees' or 'spherical-hours' or 'cartesian')";
            }
        },

        formatCartesian: function () {
            return math.trimLength(5, this.x.toFixed(2)) + math.trimLength(5, this.y.toFixed(2)) + math.trimLength(5, this.z.toFixed(2));
        },

        formatDegrees: function () {
            return "LO :" + math.trimLength(10, math.formatAngle('degrees', this.getLongitude())) + "  " +
                "LAT:" + math.trimLength(10, math.formatAngle('degrees-latitude', this.getLatitude())) + "  " +
                "RAD:" + math.trimLength(5, this.getDistance().toFixed(2));
        },

        formatDegreesSimple: function () {
            return "LO :" + math.trimLength(10, math.formatAngle('degrees-simple', this.getLongitude())) + "  " +
                "LAT:" + math.trimLength(10, math.formatAngle('degrees-simple-latitude', this.getLatitude())) + "  " +
                "RAD:" + math.trimLength(5, this.getDistance().toFixed(2));
        },

        formatHours: function (options) {
            return "REC:" + math.trimLength(10, math.formatAngle('hours', this.getLongitude())) + "  " +
                "DEC:" + math.trimLength(10, math.formatAngle('degrees-latitude', this.getLatitude())) +
                (options.distance ? "  RAD:" + math.trimLength(5, this.getDistance().toFixed(2)) : "");
        },

        addVector: function (v) {
            return math.vectorList().addVector(this).addVector(v);
        }
    });

    math.nullVector = math.vector(0, 0, 0);

    math.vectorList = core.createClass({

        init: function () {
            this.vectors = [];
        },

        addVector: function (v) {

            // new vector list
            var result = math.vectorList();

            // append own vectors
            result.vectors.push.apply(result.vectors, this.vectors);

            // append v vectors
            if (v instanceof math.vector) {
                result.vectors.push(v);
            } else if (v instanceof math.vectorList) {
                result.vectors.push.apply(result.vectors, v.vectors);
            }

            return result;
        },

        rotX: function (alpha) {
            var result = math.vectorList();
            result.vectors = this.vectors.map(function (v) {
                return v.rotX(alpha);
            });
            return result;
        },

        rotY: function (alpha) {
            var result = math.vectorList();
            result.vectors = this.vectors.map(function (v) {
                return v.rotY(alpha);
            });
            return result;
        },

        rotZ: function (alpha) {
            var result = math.vectorList();
            result.vectors = this.vectors.map(function (v) {
                return v.rotZ(alpha);
            });
            return result;
        }
    });

    return math;
});