/* global console, define, window, $, document */
define(["./astro.core", "./astro.data", "./astro.math", "./astro.calc", "./astro.map", "jquery", "raphael"], function (core, data, math, calc, map, $, raphael) {

    "use strict";

    // =========================================================================
    // packages
    // =========================================================================
    var module = {};

    // =========================================================================
    // planet lister
    // =========================================================================

    module.Lister = core.createClass({


        init: function (parentNode) {
            var self = this;
            this.parentNode = parentNode;

            // date input fields
            this.parentNode.append("<h2>Input</h2>");
            this.inputTableNode = $("<table></table>");
            this.parentNode.append(this.inputTableNode);
            var date = calc.currentDate();
            this.dayInputNode = this.createInputField(this.inputTableNode, "day", "" + date.day, "UTC (1..31)");
            this.monthInputNode = this.createInputField(this.inputTableNode, "month", "" + date.month, "UTC (1..12)");
            this.yearInputNode = this.createInputField(this.inputTableNode, "year", "" + date.year, "UTC");
            this.hourInputNode = this.createInputField(this.inputTableNode, "hour", "" + date.hour, "UTC (0..23, UTC=MEZ-1 or UTC=MESZ-2)");
            this.minuteInputNode = this.createInputField(this.inputTableNode, "minute", "" + date.minute, "UTC (0..59)");

            // geo input fields
            this.geoLatitudeInputNode = this.createInputField(this.inputTableNode, "latitude", 52, "degrees (berlin = 52 )");
            this.geoLongitudeInputNode = this.createInputField(this.inputTableNode, "longitude", 13, "degrees (<0 for west, berlin = 13)");

            // calculate button
            this.calculateButtonNode = $("<button>Calculate</button>");
            this.calculateButtonNode.click(function () {
                self.calculate();
            });
            this.parentNode.append(this.calculateButtonNode);

            // placeholder time output
            this.timeNode = $("<div class='calcsection'> </div>");
            this.parentNode.append(this.timeNode);

            // placeholder planet list output
            this.listNode = $("<div class='calcsection'></div>");
            this.parentNode.append(this.listNode);

            // placeholder map output
            this.mapNode = $("<div class='calcsection'></div>");
            this.parentNode.append(this.mapNode);

            self.calculate();

        },

        calculate: function () {

            var day = parseInt(this.dayInputNode.val(), 10);
            var month = parseInt(this.monthInputNode.val(), 10);
            var year = parseInt(this.yearInputNode.val(), 10);

            var hour = parseInt(this.hourInputNode.val(), 10);
            var minute = parseInt(this.minuteInputNode.val(), 10);
            var ut = hour + minute / 60;

            this.latitude = parseFloat(this.geoLatitudeInputNode.val()) * Math.PI / 180;
            this.longitude = parseFloat(this.geoLongitudeInputNode.val()) * Math.PI / 180;

            this.julianDate = calc.julianDate(day, month, year, ut);
            this.siderealTime0 = calc.siderealTime(this.julianDate, 0);
            this.siderealTime = calc.siderealTime(this.julianDate, this.longitude);

            this.calculatePlanets();

            this.draw();

        },

        calculatePlanets: function () {
            var planets = calc.planets(this.julianDate, this.latitude, this.longitude);
            this.planets = planets.list;
        },

        createInputField: function (parentNode, label, value, comment) {

            var containerNode = $("<tr></tr>");
            parentNode.append(containerNode);

            var labelNode = $("<td><span>" + label + "</span></td>");
            containerNode.append(labelNode);

            var valueNode = $("<td></td>");
            containerNode.append(valueNode);
            var inputNode = $("<input type='text'></input>");
            if (value) {
                inputNode.attr("value", value);
            }
            valueNode.append(inputNode);

            if (comment) {
                var commentNode = $("<td>" + comment + "</td>");
                containerNode.append(commentNode);
            }
            return inputNode;
        },

        draw: function () {
            this.drawTime();
            this.drawPlanetList();
            this.drawMap();
        },

        drawMap: function () {

            // clear old map
            var self = this;
            this.mapNode.empty();

            // heading
            this.mapNode.append("<h2>Starmap</h2>");
            this.mapNode.append("<p> red: ellipsoidal orbit <br>blue: circular orbit</p>");

            // map container
            var mapContainerNode = $("<div></div>");
            this.mapNode.append(mapContainerNode);

            // create map
            var width = Math.min(800, mapContainerNode.width());
            var paper = raphael(mapContainerNode[0], width, width);
            var mymap = map.simpleMap({
                paper: paper,
                tranformation: function (obj) {
                    return obj.rotZ(-self.siderealTime).rotY(-Math.PI / 2 + self.latitude);
                }
            });

            // add planets
            for (var i = 0; i < this.planets.length; ++i) {
                var planet = this.planets[i];
                if (planet.orbitalElements.data.id === 'earth') {
                    continue;
                }

                planet.geocentricEquatorial.title = planet.orbitalElements.name;
                mymap.addPlanet(planet.geocentricEquatorial);

                planet.geocentricEquatorialApprox.title = planet.orbitalElements.name;
                planet.geocentricEquatorialApprox.color = "#0000ff";
                mymap.addPlanet(planet.geocentricEquatorialApprox);

            }

            // draw
            mymap.draw();

        },

        drawTime: function () {
            this.timeNode.empty();
            this.timeNode.append("<h2>Time</h2>");
            var containerNode = $("<table></table>");
            this.timeNode.append(containerNode);
            containerNode.append($("<tr><td>Julian Date</td><td>" + this.julianDate + "</td></tr>"));
            containerNode.append($("<tr><td>Local Mean Sidereal Time</td><td>" + math.formatAngle("hours", this.siderealTime) + "</td></tr>"));
        },

        drawPlanetList: function () {

            // clear
            this.listNode.empty();

            // kepler equation
            var planet;
            this.listNode.append("<h2>Ephemerides Assuming Ellipsoidal Orbit</h2>");
            var table = core.Table();
            table.addRow(['name', 'orbital', 'heliocentric', 'geocentric', 'equatorial', /*'horizontal',*/ 'horizontal south', 'horizontal north']);
            for (var i = 0; i < this.planets.length; ++i) {
                planet = this.planets[i];
                table.addRow($.map([planet.orbitalElements.data.name,
                              planet.heliocentricOrbital.format('spherical-degrees'),
                              planet.heliocentricEcliptical.format('spherical-degrees'),
                              planet.geocentricEcliptical.format('spherical-degrees'),
                              planet.geocentricEquatorial.format('spherical-hours'),
                        //      planet.horizontalMath.format('spherical-degrees'),
                              planet.horizontalSouth.format('spherical-degrees'),
                              planet.horizontalNorth.format('spherical-degrees')], function (n) {
                    return $("<pre> " + n + " </pre>");
                }));
            }
            //table.mirror();
            var tableNode = table.render();
            tableNode.addClass("planettable");
            this.listNode.append(tableNode);

            // circle approximation
            this.listNode.append("<h2>Ephemerides Assuming Circular Orbit</h2>");
            table = core.Table();
            table.addRow(['name', 'heliocentric', 'geocentric', 'equatorial']);
            for (i = 0; i < this.planets.length; ++i) {
                planet = this.planets[i];
                table.addRow($.map([planet.orbitalElements.data.name,
                              planet.heliocentricEclipticalApprox.format('spherical-degrees'),
                              planet.geocentricEclipticalApprox.format('spherical-degrees'),
                              planet.geocentricEquatorialApprox.format('spherical-hours')], function (n) {
                    return $("<pre> " + n + " </pre>");
                }));
            }
            //table.mirror();
            tableNode = table.render();
            tableNode.addClass("planettable");
            this.listNode.append(tableNode);

        },

        planetRow: function (obj) {

            var cell = function (vector, format) {
                var content = "-";
                if (vector) {
                    content = vector.format(format);
                }
                return $("<td><pre>" + content + "</pre></td>");
            };

            var rowNode = $("<tr></tr>");

            // name
            var cellNode = $("<td></td>");
            rowNode.append(cellNode);
            cellNode.append(obj.orbitalElements.name);

            // coordinate
            rowNode.append(cell(obj.heliocentricOrbital, 'spherical-degrees'));
            rowNode.append(cell(obj.heliocentricEcliptical, 'spherical-degrees'));
            rowNode.append(cell(obj.geocentricEcliptical, 'spherical-degrees'));
            rowNode.append(cell(obj.geocentricEquatorial, 'spherical-hours'));
            rowNode.append(cell(obj.heliocentricEclipticalApprox, 'spherical-degrees'));
            rowNode.append(cell(obj.geocentricEclipticalApprox, 'spherical-degrees'));
            rowNode.append(cell(obj.geocentricEquatorialApprox, 'spherical-hours'));

            return rowNode;
        }

    });

    // =========================================================================
    // list planets
    // =========================================================================
    module.listPlanets = function (parentNode) {
        var lister = new module.Lister(parentNode);
    };

    return module;
});