/* global console, define, window, $, document */
define(["./astro.core", "./astro.math", "./astro.calc", "./astro.data", "./astro.map", "jquery", "raphael"], function (core, math, calc, data, map, $, raphael) {

    "use strict";

    var module = {};

    var printer = core.Printer(document.body);
    var print = core.bind(printer,printer.print);
    var printHeading = core.bind(printer,printer.printHeading);
    
    var global = {};

    // =======================================================================
    // calculate julian date    
    // =======================================================================

    var now = calc.currentDate();
    printHeading("date");
    print("year         :", now.year);
    print("month        :", now.month);
    print("day          :", now.day);
    print("hour         :", now.hour);
    print("minute       :", now.minute);

    global.julianDate = calc.julianDate(now.day, now.month, now.year, now.hour + now.minute / 60.0);
    print("julian date  :", global.julianDate);

    // =======================================================================
    // orbital coordinates of planet
    // =======================================================================

    // orbital elements
    var planet = global.planet = calc.orbitalElements(data.venus, global.julianDate);
    printHeading("orbital elements of planet (venus)");
    print("semi major axis          :", planet.semiMajorAxis);
    print("eccentricty              :", planet.eccentricity);
    print("mean anomaly             :", math.formatAngle("degrees-simple", planet.meanAnomaly));
    print("longitude ascending node :", math.formatAngle("degrees-simple", planet.longitudeAscendingNode));
    print("inclination              :", math.formatAngle("degrees-simple", planet.inclination));
    print("argument perihelion      :", math.formatAngle("degrees-simple", planet.argumentPerihelion));
    print("longitude of perihelion  :", math.formatAngle("degrees-simple", planet.longitudePerihelion));
    print("mean longitude           :", math.formatAngle("degrees-simple", planet.meanLongitude));

    // eccentric anomaly
    printHeading("orbital coordinates of planet (venus)");
    planet.eccentricAnomaly = calc.eccentricAnomaly(planet.meanAnomaly, planet.eccentricity);
    print("eccentric anomaly        : ", planet.eccentricAnomaly);

    // orbital coordinates 
    planet.orbitalCoordinates = calc.orbitalCoordinates(planet.semiMajorAxis, planet.eccentricity, planet.eccentricAnomaly);
    print("lambda                   : ", math.formatAngle("degrees-simple", planet.orbitalCoordinates.lambda));
    print("radius                   : ", planet.orbitalCoordinates.rad);

    // make vector from orbital coordinates
    planet.orbitalPosition = math.vector(0, planet.orbitalCoordinates.lambda, planet.orbitalCoordinates.rad, 'spherical');
    print("orbital coordinates (longitude/latitude)               :", planet.orbitalPosition.format("spherical-degrees"));

    // =======================================================================
    // heliocentric ecliptic coordinates of planet
    // =======================================================================
    planet.heliocentricEclipticPosition = planet.orbitalPosition
        .rotZ(planet.argumentPerihelion)
        .rotX(planet.inclination)
        .rotZ(planet.longitudeAscendingNode);
    printHeading("heliocentric ecliptic coordinates of planet (venus)");
    print("heliocentric ecliptic coordinates (longitude/latitude) :", planet.heliocentricEclipticPosition.format("spherical-degrees"));

    // =======================================================================
    // orbital coordinates of earth
    // =======================================================================

    // orbital elements
    var earth = global.earth = calc.orbitalElements(data.earth, global.julianDate);
    printHeading("orbital elements of earth");
    print("semi major axis          :", earth.semiMajorAxis);
    print("eccentricty              :", earth.eccentricity);
    print("mean anomaly             :", math.formatAngle("degrees-simple", earth.meanAnomaly));
    print("longitude ascending node :", math.formatAngle("degrees-simple", earth.longitudeAscendingNode));
    print("inclination              :", math.formatAngle("degrees-simple", earth.inclination));
    print("argument perihelion      :", math.formatAngle("degrees-simple", earth.argumentPerihelion));
    print("longitude of perihelion  :", math.formatAngle("degrees-simple", earth.longitudePerihelion));
    print("mean longitude           :", math.formatAngle("degrees-simple", earth.meanLongitude));

    // eccentric anomaly
    earth.eccentricAnomaly = calc.eccentricAnomaly(earth.meanAnomaly, earth.eccentricity);
    print("eccentric anomaly        : ", earth.eccentricAnomaly);

    // orbital position
    earth.orbitalCoordinates = calc.orbitalCoordinates(earth.semiMajorAxis, earth.eccentricity, earth.eccentricAnomaly);
    print("lambda                   : ", math.formatAngle("degrees-simple", earth.orbitalCoordinates.lambda));
    print("radius                   : ", earth.orbitalCoordinates.rad);

    // make vector from orbital coordinates
    earth.orbitalPosition = math.vector(0, earth.orbitalCoordinates.lambda, earth.orbitalCoordinates.rad, 'spherical');
    print("orbital coordinates (longitude/latitude)                :", earth.orbitalPosition.format("spherical-degrees"));

    // =======================================================================
    // heliocentric ecliptic coordinates of earth
    // =======================================================================
    earth.heliocentricEclipticPosition = earth.orbitalPosition
        .rotZ(earth.argumentPerihelion)
        .rotX(earth.inclination)
        .rotZ(earth.longitudeAscendingNode);
    printHeading("heliocentric ecliptic coordinates of earth");
    print("heliocentric ecliptic coordinates (longitude/latitude)  :", earth.heliocentricEclipticPosition.format("spherical-degrees"));

    // =======================================================================
    // geocentric ecliptic coordinates of planet
    // =======================================================================
    planet.geocentricEclipticPosition = planet.heliocentricEclipticPosition.subtract(earth.heliocentricEclipticPosition);
    printHeading("geocentric eclpitic coordinates of planet (venus)");
    print("geocentric eclpitic coordinates (longitude/latitude)    :", planet.geocentricEclipticPosition.format('spherical-degrees'));

    // =======================================================================
    // geocentric equatorial coordinates of planet
    // =======================================================================
    planet.geocentricEquatorialPosition = planet.geocentricEclipticPosition.rotX(calc.eclipticAngle(global.julianDate));
    printHeading("geocentric equatorial coordinates of planet (venus)");
    print("geocentric equatorial coordinates (longitude/latitude)  :", planet.geocentricEquatorialPosition.format('spherical-hours'));

    // =======================================================================
    // map
    // =======================================================================
    printHeading("map");
    var mapNode = $("<div></div>");
    $("body").append(mapNode);
    var mymap = map.simpleMap({
        paper: raphael(mapNode[0], 500, 500)
    });

    planet.color="red";
    planet.title="Planet";
    mymap.addPlanet(planet.geocentricEquatorialPosition);

    // draw
    mymap.draw();

    return module;
});