/* global console, define, window, $, document */
define(["./astro.core", "./astro.data", "./astro.math", "./astro.calc", "./astro.map", "jquery", "flot", "flot.time"], function (core, data, math, calc, map, $, flot, flottime) {

    "use strict";

    // =======================================================================
    // plot helper
    // =======================================================================
    var plotHelper = core.createClass({

        init: function () {
            this.plotMap = {};
        },

        addPoint: function (id, point) {
            var plotData = this.plotMap[id];
            if (plotData === undefined) {
                plotData = [];
                this.plotMap[id] = plotData;
            }
            plotData.push(point);
        },

        addElongation: function (id, point) {
            var last = this.getLastPoint(id);
            if (last) {
                if (last[1] * point[1] < 0 && Math.abs(point[1]) > 90 && Math.abs(last[1]) > 90) {
                    this.addPoint(id, null);
                }
            }
            this.addPoint(id, point);
        },

        getLastPoint: function (id) {
            var plotData = this.plotMap[id];
            if (plotData === undefined) {
                return null;
            }
            if (plotData.length > 0) {
                return plotData[plotData.length - 1];
            } else {
                return null;
            }
        },

        getPlots: function () {
            var allPlots = [];
            for (var id in this.plotMap) {
                allPlots.push({
                    label: id,
                    data: this.plotMap[id]
                });
            }
            return allPlots;
        },

        draw: function (node) {

            var allPlots = [];
            for (var id in this.plotMap) {
                allPlots.push({
                    label: id,
                    data: this.plotMap[id]
                });
            }

            $.plot(node, allPlots, {
                series: {
                    lines: {
                        show: true
                    },
                    points: {
                        show: false
                    }
                },
                xaxis: {
                    mode: "time",
                    minTickSize: [1, "month"],
                }
            });
        }

    });

    // =======================================================================
    // plotter
    // =======================================================================
    var plotter = core.createClass({

        init: function (options) {

            // init
            var self = this;
            self.options = options;

            // create main container 
            self.container = $("<div></div>");
            self.options.parentNode.append(self.container);

            // toggle button for dimensions
            self.toggleDimensionsButton = $("<button>show configuration</button>");
            self.container.append(self.toggleDimensionsButton);
            self.toggleDimensionsButton.click(function () {
                self.dimensionsContainer.toggle();
            });

            // container for all dimensions
            self.dimensionsContainer = $("<div></div>");
            self.container.append(self.dimensionsContainer);
            $.each(options.dimensions, function (index, dimension) {
                self.createDimensionSelection(dimension);
            });

            // checkbox for parameter plot
            self.parameterPlotCheckbox = $("<input type='checkbox'><span>parameter plot</span>");
            self.dimensionsContainer.append(self.parameterPlotCheckbox);

            // plot button
            var plotButton = $("<button>plot</button>");
            self.container.append(plotButton);
            plotButton.click(function () {
                self.plot();
            });

            // plot area
            self.plotContainer = $("<div style='width:1400px;height:800px'></div>");
            self.container.append(self.plotContainer);
        },

        parameterPlot: function () {
            var self = this;
            self.options.x.reset();
            var plots = plotHelper();

            var props = [];
            for (var j = 0; j < self.options.dimensions[1].values.length; ++j) {
                var value = self.options.dimensions[1].values[j];
                if (!self.options.dimensions[1].checkboxes[value.id].prop('checked')) {
                    continue;
                }
                props.push(value);
            }

            for (; self.options.x.hasNext(); self.options.x.next()) {
                var x = self.options.x.value();
                var y = self.options.f(x);
                for (var i = 0; i < self.options.dimensions[0].values.length; ++i) {
                    var planetId = self.options.dimensions[0].values[i].id;
                    if (!self.options.dimensions[0].checkboxes[planetId].prop('checked')) {
                        continue;
                    }
                    var xv = y[planetId][props[0].id];
                    var yv = y[planetId][props[1].id];
                    plots.addPoint(planetId, [xv, yv]);
                }
            }

            $.plot(this.plotContainer, plots.getPlots(), {
                series: {
                    lines: {
                        show: true
                    },
                    points: {
                        show: false
                    }
                },
                xaxis: {}
            });

        },

        plot: function () {

            this.plotContainer.empty();
            var self = this;
            if (self.parameterPlotCheckbox.prop('checked')) {
                self.parameterPlot();
                return;
            }
            var plots = plotHelper();
            var paths = [];
            self.collectDimensionValues(0, [], paths);
            self.options.x.reset();
            for (; self.options.x.hasNext(); self.options.x.next()) {
                var x = self.options.x.value();
                var y = self.options.f(x);
                for (var i = 0; i < paths.length; ++i) {
                    var path = paths[i];
                    var obj = y;
                    for (var j = 0; j < path.length; ++j) {
                        var key = path[j];
                        obj = obj[key];
                    }
                    plots.addElongation(path[0] + " " + path[1].slice(5), [self.options.x.plotValue(), obj]);
                }
            }
            plots.draw(this.plotContainer);
        },

        collectDimensionValues: function (index, path, result) {
            if (index >= this.options.dimensions.length) {
                result.push(path);
                return;
            }
            var dimension = this.options.dimensions[index];
            for (var i = 0; i < dimension.values.length; ++i) {
                var value = dimension.values[i];
                if (!dimension.checkboxes[value.id].prop('checked')) {
                    continue;
                }
                var newPath = path.slice();
                newPath.push(value.id);
                this.collectDimensionValues(index + 1, newPath, result);
            }
        },

        createDimensionSelection: function (dimension) {
            var self = this;
            var container = $("<div></div>");
            self.dimensionsContainer.append(container);
            dimension.checkboxes = {};
            container.append($("<h2>" + dimension.id + "</h2>"));
            $.each(dimension.values, function (index, value) {

                // set id and label
                var id = value.id;
                var label = value.label;

                // create checkbox
                var checkbox = $("<input type='checkbox'>");
                container.append(checkbox);
                container.append($("<span>" + label + "</span><br>"));

                // register checkbox
                dimension.checkboxes[id] = checkbox;
            });
        }

    });

    var yearTimeRange = core.createClass({

        init: function (year) {
            this.year = year;
            this.day = 1;
            this.month = 1;
        },

        reset: function () {
            this.day = 1;
            this.month = 1;
        },

        value: function () {
            return calc.julianDate(this.day, this.month, this.year, 0);
        },

        plotValue: function () {
            return new Date(this.year, this.month - 1, this.day).getTime();
        },

        next: function () {
            if (this.day < calc.daysOfMonth(this.month)) {
                this.day += 1;
                return;
            }
            if (this.month < 12) {
                this.day = 1;
                this.month += 1;
                return;
            }
            throw "end of year";
        },

        hasNext: function () {
            if (this.day < calc.daysOfMonth(this.month)) {
                return true;
            }
            if (this.month < 12) {
                return true;
            }
            return false;
        }

    });

    // =======================================================================
    // main module
    // =======================================================================
    var module = {

        descriptionMap: {
            //     helioX : 'helicoentric X'
        },

        getDescription: function (attribute) {
            var descr = this.descriptionMap[attribute];
            if (!descr) descr = attribute;
            return descr;
        },

        getParameters: function () {
            var parameters = [];
            var planets = calc.planets(calc.julianDate(1, 1, 2000, 0));
            for (var attribute in planets.venus) {
                if (attribute.slice(0, 5) === 'plot_') {
                    parameters.push({
                        id: attribute,
                        label: this.getDescription(attribute.slice(5))
                    });
                }
            }
            return parameters;
        },

        drawVisibility: function (node) {

            var parameters = this.getParameters();
            var addButton = $("<button>Add Plot</button>");
            node.append(addButton);
            addButton.click(function () {

                plotter({
                    parentNode: node,
                    x: yearTimeRange(2014),
                    f: function (julianDate) {
                        return calc.planets(julianDate);
                    },
                    dimensions: [{
                            id: 'planet',
                            values: $.map(['venus', 'mars', 'jupiter', 'saturn', 'sun', 'earth'], function (id) {
                                return {
                                    id: id,
                                    label: id
                                };
                            })
                    },
                        {
                            id: 'value',
                            values: parameters
                    }]
                });

            });
        }

    };

    return module;
});